@@ -207,6 +207,7 @@ read_delim_arrow <- function(file,
     }
     on.exit(file$close())
   }
+
   reader <- CsvTableReader$create(
     file,
     read_options = read_options,
@@ -326,6 +327,18 @@ CsvTableReader$create <- function(file,
                                   ...) {
   assert_is(file, "InputStream")
 
+  if (is.list(read_options)) {
+    read_options <- do.call(CsvReadOptions$create, read_options)
+  }
+
+  if (is.list(parse_options)) {
+    parse_options <- do.call(CsvParseOptions$create, parse_options)
+  }
+
+  if (is.list(convert_options)) {
+    convert_options <- do.call(CsvConvertOptions$create, convert_options)
+  }
+
   if (!(tolower(read_options$encoding) %in% c("utf-8", "utf8"))) {
     file <- MakeReencodeInputStream(file, read_options$encoding)
   }
@@ -418,6 +431,8 @@ CsvTableReader$create <- function(file,
 #' The `CsvWriteOptions$create()` factory method takes the following arguments:
 #' - `include_header` Whether to write an initial header line with column names
 #' - `batch_size` Maximum number of rows processed at a time. Default is 1024.
+#' - `null_string` The string to be written for null values. Must not contain
+#'   quotation marks. Default is an empty string (`""`).
 #'
 #' @section Active bindings:
 #'
@@ -455,30 +470,37 @@ CsvReadOptions$create <- function(use_threads = option_use_threads(),
   options
 }
 
-readr_to_csv_write_options <- function(include_header,
-                                       batch_size = 1024L) {
-  assert_that(is_integerish(batch_size, n = 1, finite = TRUE), batch_size > 0)
-  assert_that(is.logical(include_header))
+readr_to_csv_write_options <- function(include_header = TRUE,
+                                       batch_size = 1024L,
+                                       na = "") {
   CsvWriteOptions$create(
     include_header = include_header,
-    batch_size = as.integer(batch_size)
+    batch_size = batch_size,
+    null_string = na
   )
 }
 
 #' @rdname CsvReadOptions
 #' @export
 CsvWriteOptions <- R6Class("CsvWriteOptions", inherit = ArrowObject)
-CsvWriteOptions$create <- function(include_header = TRUE, batch_size = 1024L) {
+CsvWriteOptions$create <- function(include_header = TRUE, batch_size = 1024L, null_string = "") {
   assert_that(is_integerish(batch_size, n = 1, finite = TRUE), batch_size > 0)
+  assert_that(is.logical(include_header))
+  assert_that(is.character(null_string))
+  assert_that(!is.na(null_string))
+  assert_that(length(null_string) == 1)
+  assert_that(!grepl('"', null_string), msg = "na argument must not contain quote characters.")
+
   csv___WriteOptions__initialize(
     list(
       include_header = include_header,
-      batch_size = as.integer(batch_size)
+      batch_size = as.integer(batch_size),
+      null_string = as.character(null_string)
     )
   )
 }
 
-readr_to_csv_read_options <- function(skip, col_names, col_types) {
+readr_to_csv_read_options <- function(skip = 0, col_names = TRUE) {
   if (isTRUE(col_names)) {
     # C++ default to parse is 0-length string array
     col_names <- character(0)
@@ -665,6 +687,8 @@ readr_to_csv_convert_options <- function(na,
 #' @param col_names identical to `include_header`. Specify this or
 #'     `include_headers`, not both.
 #' @param batch_size Maximum number of rows processed at a time. Default is 1024.
+#' @param na value to write for NA values. Must not contain quote marks. Default
+#'     is `""`.
 #' @param write_options see [file reader options][CsvWriteOptions]
 #' @param ... additional parameters
 #'
@@ -682,6 +706,7 @@ write_csv_arrow <- function(x,
                             include_header = TRUE,
                             col_names = NULL,
                             batch_size = 1024L,
+                            na = "",
                             write_options = NULL,
                             ...) {
   unsupported_passed_args <- names(list(...))
@@ -723,7 +748,8 @@ write_csv_arrow <- function(x,
   if (is.null(write_options)) {
     write_options <- readr_to_csv_write_options(
       include_header = include_header,
-      batch_size = batch_size
+      batch_size = batch_size,
+      na = na
     )
   }
 
