@@ -21,6 +21,7 @@ register_bindings_datetime <- function() {
   register_bindings_datetime_utility()
   register_bindings_datetime_components()
   register_bindings_datetime_conversion()
+  register_bindings_datetime_timezone()
   register_bindings_duration()
   register_bindings_duration_constructor()
   register_bindings_duration_helpers()
@@ -45,7 +46,7 @@ register_bindings_datetime_utility <- function() {
         c(valid_time64_units, valid_time32_units)
       )
 
-      output <- build_expr(
+      output <- Expression$create(
         "strptime",
         x,
         options =
@@ -66,7 +67,7 @@ register_bindings_datetime_utility <- function() {
       # if we want to cast to a different timezone, we can only do it for
       # timezone-aware timestamps, not for timezone-naive ones
       if (!is.null(tz)) {
-        output <- build_expr(
+        output <- Expression$create(
           "assume_timezone",
           output,
           options =
@@ -152,7 +153,7 @@ register_bindings_datetime_utility <- function() {
   })
 
   register_binding("lubridate::date", function(x) {
-    build_expr("cast", x, options = list(to_type = date32()))
+    cast(x, date32())
   })
 }
 
@@ -202,7 +203,7 @@ register_bindings_datetime_components <- function() {
       # days from epoch (we multiply by 28 to be able to later extract the
       # month with label) - NB this builds a false date (to be used by strftime)
       # since we only know and care about the month
-      x <- build_expr("cast", x * 28L, options = cast_options(to_type = date32()))
+      x <- cast(x * 28L, date32())
     }
 
     if (label) {
@@ -211,18 +212,18 @@ register_bindings_datetime_components <- function() {
       } else {
         format <- "%B"
       }
-      return(build_expr("strftime", x, options = list(format = format, locale = check_time_locale(locale))))
+      return(Expression$create("strftime", x, options = list(format = format, locale = check_time_locale(locale))))
     }
 
-    build_expr("month", x)
+    Expression$create("month", x)
   })
 
   register_binding("lubridate::qday", function(x) {
     # We calculate day of quarter by flooring timestamp to beginning of quarter and
     # calculating days between beginning of quarter and timestamp/date in question.
     # Since we use one one-based numbering we add one.
-    floored_x <- build_expr("floor_temporal", x, options = list(unit = 9L))
-    build_expr("days_between", floored_x, x) + Expression$scalar(1L)
+    floored_x <- Expression$create("floor_temporal", x, options = list(unit = 9L))
+    Expression$create("days_between", floored_x, x) + Expression$scalar(1L)
   })
 
   register_binding("lubridate::am", function(x) {
@@ -234,11 +235,11 @@ register_bindings_datetime_components <- function() {
   })
   register_binding("lubridate::tz", function(x) {
     if (!call_binding("is.POSIXct", x)) {
-      abort(
+      arrow_not_supported(
         paste0(
           "timezone extraction for objects of class `",
           infer_type(x)$ToString(),
-          "` not supported in Arrow"
+          "`"
         )
       )
     }
@@ -247,7 +248,7 @@ register_bindings_datetime_components <- function() {
   })
   register_binding("lubridate::semester", function(x, with_year = FALSE) {
     month <- call_binding("month", x)
-    semester <- call_binding("if_else", month <= 6, 1L, 2L)
+    semester <- Expression$create("if_else", month <= 6, 1L, 2L)
     if (with_year) {
       year <- call_binding("year", x)
       return(year + semester / 10)
@@ -267,7 +268,6 @@ register_bindings_datetime_conversion <- function() {
              min = 0L,
              sec = 0,
              tz = "UTC") {
-
       # ParseTimestampStrptime currently ignores the timezone information (ARROW-12820).
       # Stop if tz other than 'UTC' is provided.
       if (tz != "UTC") {
@@ -275,7 +275,7 @@ register_bindings_datetime_conversion <- function() {
       }
 
       x <- call_binding("str_c", year, month, day, hour, min, sec, sep = "-")
-      build_expr("strptime", x, options = list(format = "%Y-%m-%d-%H-%M-%S", unit = 0L))
+      Expression$create("strptime", x, options = list(format = "%Y-%m-%d-%H-%M-%S", unit = 0L))
     },
     notes = "only supports UTC (default) timezone"
   )
@@ -284,7 +284,7 @@ register_bindings_datetime_conversion <- function() {
                                                     month = 1L,
                                                     day = 1L) {
     x <- call_binding("make_datetime", year, month, day)
-    build_expr("cast", x, options = cast_options(to_type = date32()))
+    cast(x, date32())
   })
 
   register_binding("base::ISOdatetime", function(year,
@@ -294,7 +294,6 @@ register_bindings_datetime_conversion <- function() {
                                                  min,
                                                  sec,
                                                  tz = "UTC") {
-
     # NAs for seconds aren't propagated (but treated as 0) in the base version
     sec <- call_binding(
       "if_else",
@@ -339,7 +338,8 @@ register_bindings_datetime_conversion <- function() {
       # => we only cast when we want the behaviour of the base version or when
       # `tz` is set (i.e. not NULL)
       if (call_binding("is.POSIXct", x)) {
-        x <- build_expr("cast", x, options = cast_options(to_type = timestamp(timezone = tz)))
+        unit <- if (inherits(x, "Expression")) x$type()$unit() else "s"
+        x <- cast(x, timestamp(unit = unit, timezone = tz))
       }
 
       binding_as_date(
@@ -366,7 +366,8 @@ register_bindings_datetime_conversion <- function() {
     # => we only cast when we want the behaviour of the base version or when
     # `tz` is set (i.e. not NULL)
     if (call_binding("is.POSIXct", x) && !is.null(tz)) {
-      x <- build_expr("cast", x, options = cast_options(to_type = timestamp(timezone = tz)))
+      unit <- if (inherits(x, "Expression")) x$type()$unit() else "s"
+      x <- cast(x, timestamp(unit = unit, timezone = tz))
     }
     binding_as_date(
       x = x,
@@ -378,66 +379,146 @@ register_bindings_datetime_conversion <- function() {
   register_binding("lubridate::as_datetime", function(x,
                                                       origin = "1970-01-01",
                                                       tz = "UTC",
-                                                      format = NULL) {
+                                                      format = NULL,
+                                                      unit = "ns") {
+    # Arrow uses unit for time parsing, as_datetime() does not.
+    unit <- make_valid_time_unit(
+      unit,
+      c(valid_time64_units, valid_time32_units)
+    )
+
+    if (call_binding("is.integer", x)) {
+      x <- cast(x, int64())
+    }
+
     if (call_binding("is.numeric", x)) {
+      multiple <- Expression$create("power_checked", 1000L, unit)
       delta <- call_binding("difftime", origin, "1970-01-01")
-      delta <- build_expr("cast", delta, options = cast_options(to_type = int64()))
-      x <- build_expr("cast", x, options = cast_options(to_type = int64()))
-      x <- build_expr("+", x, delta)
+      delta <- cast(delta, int64())
+      delta <- Expression$create("multiply_checked", delta, multiple)
+      x <- Expression$create("multiply_checked", x, multiple)
+      x <- cast(x, int64())
+      x <- Expression$create("add_checked", x, delta)
     }
 
     if (call_binding("is.character", x) && !is.null(format)) {
-      # unit = 0L is the identifier for seconds in valid_time32_units
-      x <- build_expr(
+      x <- Expression$create(
         "strptime",
         x,
-        options = list(format = format, unit = 0L, error_is_null = TRUE)
+        options = list(format = format, unit = unit, error_is_null = TRUE)
       )
     }
-    output <- build_expr("cast", x, options = cast_options(to_type = timestamp()))
-    build_expr("assume_timezone", output, options = list(timezone = tz))
+    output <- cast(x, timestamp(unit = unit))
+    Expression$create("assume_timezone", output, options = list(timezone = tz))
   })
 
   register_binding("lubridate::decimal_date", function(date) {
-    y <- build_expr("year", date)
+    y <- Expression$create("year", date)
     start <- call_binding("make_datetime", year = y, tz = "UTC")
     sofar <- call_binding("difftime", date, start, units = "secs")
-    total <- call_binding(
+    total <- Expression$create(
       "if_else",
-      build_expr("is_leap_year", date),
+      Expression$create("is_leap_year", date),
       Expression$scalar(31622400L), # number of seconds in a leap year (366 days)
       Expression$scalar(31536000L) # number of seconds in a regular year (365 days)
     )
-    y + sofar$cast(int64()) / total
+    y + cast(sofar, int64()) / total
   })
 
   register_binding("lubridate::date_decimal", function(decimal, tz = "UTC") {
-    y <- build_expr("floor", decimal)
+    y <- Expression$create("floor", decimal)
 
     start <- call_binding("make_datetime", year = y, tz = tz)
-    seconds <- call_binding(
+    seconds <- Expression$create(
       "if_else",
-      build_expr("is_leap_year", start),
+      Expression$create("is_leap_year", start),
       Expression$scalar(31622400L), # number of seconds in a leap year (366 days)
       Expression$scalar(31536000L) # number of seconds in a regular year (365 days)
     )
 
     fraction <- decimal - y
-    delta <- build_expr("floor", seconds * fraction)
+    delta <- Expression$create("floor", seconds * fraction)
     delta <- make_duration(delta, "s")
     start + delta
   })
 }
 
+register_bindings_datetime_timezone <- function() {
+  register_binding(
+    "lubridate::force_tz",
+    function(time, tzone = "", roll_dst = c("error", "post")) {
+      if (length(roll_dst) == 1L) {
+        roll_dst <- c(roll_dst, roll_dst)
+      } else if (length(roll_dst) != 2L) {
+        arrow_not_supported("`roll_dst` must be 1 or 2 items long; other lengths")
+      }
+
+      nonexistent <- switch(
+        roll_dst[1],
+        "error" = 0L,
+        "boundary" = 2L,
+        arrow_not_supported("`roll_dst` value must be 'error' or 'boundary' for non-existent times; other values")
+      )
+
+      ambiguous <- switch(
+        roll_dst[2],
+        "error" = 0L,
+        "pre" = 1L,
+        "post" = 2L,
+        arrow_not_supported("`roll_dst` value must be 'error', 'pre', or 'post' for non-existent times")
+      )
+
+      if (identical(tzone, "")) {
+        tzone <- Sys.timezone()
+      }
+
+      if (!inherits(time, "Expression")) {
+        time <- Expression$scalar(time)
+      }
+
+      # Non-UTC timezones don't work here and getting them to do so was too
+      # hard to do in the initial PR because there is no way in Arrow to
+      # "unapply" a UTC offset (i.e., the reverse of assume_timezone).
+      if (!time$type()$timezone() %in% c("", "UTC")) {
+        arrow_not_supported("`time` with a non-UTC timezone")
+      }
+
+      # Remove timezone if needed
+      current_unit <- time$type()$unit()
+      time <- cast(time, timestamp(current_unit, ""))
+
+      # Add timezone
+      Expression$create(
+        "assume_timezone",
+        time,
+        options = list(
+          timezone = tzone,
+          nonexistent = nonexistent,
+          ambiguous = ambiguous
+        )
+      )
+    },
+    notes = c(
+      "Timezone conversion from non-UTC timezone not supported;",
+      "`roll_dst` values of 'error' and 'boundary' are supported for nonexistent times,",
+      "`roll_dst` values of 'error', 'pre', and 'post' are supported for ambiguous times."
+    )
+  )
+
+  register_binding("lubridate::with_tz", function(time, tzone = "") {
+    if (tzone == "") {
+      tzone <- Sys.timezone()
+    }
+    cast(time, timestamp(unit = time$type()$unit(), timezone = tzone))
+  })
+}
+
 register_bindings_duration <- function() {
   register_binding(
     "base::difftime",
-    function(time1,
-             time2,
-             tz,
-             units = "secs") {
+    function(time1, time2, tz, units = "secs") {
       if (units != "secs") {
-        abort("`difftime()` with units other than `secs` not supported in Arrow")
+        arrow_not_supported("`difftime()` with units other than `secs`")
       }
 
       if (!missing(tz)) {
@@ -447,11 +528,11 @@ register_bindings_duration <- function() {
       # cast to timestamp if time1 and time2 are not dates or timestamp expressions
       # (the subtraction of which would output a `duration`)
       if (!call_binding("is.instant", time1)) {
-        time1 <- build_expr("cast", time1, options = cast_options(to_type = timestamp()))
+        time1 <- cast(time1, timestamp())
       }
 
       if (!call_binding("is.instant", time2)) {
-        time2 <- build_expr("cast", time2, options = cast_options(to_type = timestamp()))
+        time2 <- cast(time2, timestamp())
       }
 
       # if time1 or time2 are timestamps they cannot be expressed in "s" /seconds
@@ -460,18 +541,17 @@ register_bindings_duration <- function() {
       # https://issues.apache.org/jira/browse/ARROW-16060 is solved
       if (inherits(time1, "Expression") &&
         time1$type_id() %in% Type[c("TIMESTAMP")] && time1$type()$unit() != 2L) {
-        time1 <- build_expr("cast", time1, options = cast_options(to_type = timestamp("us")))
+        time1 <- cast(time1, timestamp("us"))
       }
 
       if (inherits(time2, "Expression") &&
         time2$type_id() %in% Type[c("TIMESTAMP")] && time2$type()$unit() != 2L) {
-        time2 <- build_expr("cast", time2, options = cast_options(to_type = timestamp("us")))
+        time2 <- cast(time2, timestamp("us"))
       }
 
       # we need to go build the subtract expression instead of `time1 - time2` to
       # prevent complaints when we try to subtract an R object from an Expression
-      subtract_output <- build_expr("-", time1, time2)
-      build_expr("cast", subtract_output, options = cast_options(to_type = duration("s")))
+      cast(call_binding("-", time1, time2), duration("s"))
     },
     notes = c(
       'only supports `units = "secs"` (the default);',
@@ -481,22 +561,20 @@ register_bindings_duration <- function() {
 
   register_binding(
     "base::as.difftime",
-    function(x,
-             format = "%X",
-             units = "secs") {
+    function(x, format = "%X", units = "secs") {
       # windows doesn't seem to like "%X"
       if (format == "%X" & tolower(Sys.info()[["sysname"]]) == "windows") {
         format <- "%H:%M:%S"
       }
 
       if (units != "secs") {
-        abort("`as.difftime()` with units other than 'secs' not supported in Arrow")
+        arrow_not_supported("`as.difftime()` with units other than 'secs'")
       }
 
       if (call_binding("is.character", x)) {
-        x <- build_expr("strptime", x, options = list(format = format, unit = 0L))
+        x <- Expression$create("strptime", x, options = list(format = format, unit = 0L))
         # we do a final cast to duration ("s") at the end
-        x <- make_duration(x$cast(time64("us")), unit = "us")
+        x <- make_duration(cast(x, time64("us")), unit = "us")
       }
 
       # numeric -> duration not supported in Arrow yet so we use int64() as an
@@ -508,10 +586,10 @@ register_bindings_duration <- function() {
         # for pure doubles
         # if we abort for all doubles, we risk erroring in cases in which
         # coercion to int64() would work
-        x <- build_expr("cast", x, options = cast_options(to_type = int64()))
+        x <- cast(x, int64())
       }
 
-      build_expr("cast", x, options = cast_options(to_type = duration(unit = "s")))
+      cast(x, duration(unit = "s"))
     },
     notes = 'only supports `units = "secs"` (the default)'
   )
@@ -520,11 +598,9 @@ register_bindings_duration <- function() {
 register_bindings_duration_constructor <- function() {
   register_binding(
     "lubridate::make_difftime",
-    function(num = NULL,
-             units = "secs",
-             ...) {
+    function(num = NULL, units = "secs", ...) {
       if (units != "secs") {
-        abort("`make_difftime()` with units other than 'secs' not supported in Arrow")
+        arrow_not_supported("`make_difftime()` with units other than 'secs'")
       }
 
       chunks <- list(...)
@@ -533,7 +609,7 @@ register_bindings_duration_constructor <- function() {
       # passed via `...` resulting in a vector of length 2 - which is virtually
       # unusable in a dplyr pipeline. Arrow errors in this situation
       if (!is.null(num) && length(chunks) > 0) {
-        abort("`make_difftime()` with both `num` and `...` not supported in Arrow")
+        arrow_not_supported("`make_difftime()` with both `num` and `...`")
       }
 
       if (!is.null(num)) {
@@ -554,22 +630,22 @@ register_bindings_duration_constructor <- function() {
 }
 
 register_bindings_duration_helpers <- function() {
-  duration_helpers_map_factory <- function(value, unit) {
+  duration_factory <- function(value, unit) {
     force(value)
     force(unit)
     function(x = 1) make_duration(x * value, unit)
   }
 
-  for (name in names(.helpers_function_map)) {
-    register_binding(
-      name,
-      duration_helpers_map_factory(
-        .helpers_function_map[[name]][[1]],
-        .helpers_function_map[[name]][[2]]
-      )
-    )
-  }
-
+  register_binding("lubridate::dminutes", duration_factory(60, "s"))
+  register_binding("lubridate::dhours", duration_factory(3600, "s"))
+  register_binding("lubridate::ddays", duration_factory(86400, "s"))
+  register_binding("lubridate::dweeks", duration_factory(604800, "s"))
+  register_binding("lubridate::dmonths", duration_factory(2629800, "s"))
+  register_binding("lubridate::dyears", duration_factory(31557600, "s"))
+  register_binding("lubridate::dseconds", duration_factory(1, "s"))
+  register_binding("lubridate::dmilliseconds", duration_factory(1, "ms"))
+  register_binding("lubridate::dmicroseconds", duration_factory(1, "us"))
+  register_binding("lubridate::dnanoseconds", duration_factory(1, "ns"))
   register_binding(
     "lubridate::dpicoseconds",
     function(x = 1) {
@@ -611,13 +687,13 @@ register_bindings_datetime_parsers <- function() {
         parse_attempts <- attempt_parsing(x, orders = orders)
       }
 
-      coalesce_output <- build_expr("coalesce", args = parse_attempts)
+      coalesce_output <- Expression$create("coalesce", args = parse_attempts)
 
       # we need this binding to be able to handle a NULL `tz`, which, in turn,
       # will be used by bindings such as `ymd()` to return a date or timestamp,
       # based on whether tz is NULL or not
       if (!is.null(tz)) {
-        build_expr("assume_timezone", coalesce_output, options = list(timezone = tz))
+        Expression$create("assume_timezone", coalesce_output, options = list(timezone = tz))
       } else {
         coalesce_output
       }
@@ -648,7 +724,7 @@ register_bindings_datetime_parsers <- function() {
         # we cast so we can mimic the behaviour of the `tz` argument in lubridate
         # "If NULL (default), a Date object is returned. Otherwise a POSIXct with
         # time zone attribute set to tz."
-        parse_x <- parse_x$cast(date32())
+        parse_x <- cast(parse_x, date32())
       }
       parse_x
     }
@@ -664,18 +740,14 @@ register_bindings_datetime_parsers <- function() {
 
   register_binding(
     "lubridate::fast_strptime",
-    function(x,
-             format,
-             tz = "UTC",
-             lt = FALSE,
-             cutoff_2000 = 68L) {
+    function(x, format, tz = "UTC", lt = FALSE, cutoff_2000 = 68L) {
       # `lt` controls the output `lt = TRUE` returns a POSIXlt (which doesn't play
       # well with mutate, for example)
       if (lt) {
         arrow_not_supported("`lt = TRUE` argument")
       }
 
-      # TODO revisit once https://issues.apache.org/jira/browse/ARROW-16596 is done
+      # TODO revisit after https://issues.apache.org/jira/browse/ARROW-16596
       if (cutoff_2000 != 68L) {
         arrow_not_supported("`cutoff_2000` != 68L argument")
       }
@@ -684,7 +756,7 @@ register_bindings_datetime_parsers <- function() {
 
       parse_attempt_expressions <- map(
         format,
-        ~ build_expr(
+        ~ Expression$create(
           "strptime",
           x,
           options = list(
@@ -695,9 +767,9 @@ register_bindings_datetime_parsers <- function() {
         )
       )
 
-      coalesce_output <- build_expr("coalesce", args = parse_attempt_expressions)
+      coalesce_output <- Expression$create("coalesce", args = parse_attempt_expressions)
 
-      build_expr("assume_timezone", coalesce_output, options = list(timezone = tz))
+      Expression$create("assume_timezone", coalesce_output, options = list(timezone = tz))
     },
     notes = "non-default values of `lt` and `cutoff_2000` not supported"
   )
