@@ -293,7 +293,8 @@ std::shared_ptr<fs::S3FileSystem> fs___S3FileSystem__create(
     std::string session_name = "", std::string external_id = "", int load_frequency = 900,
     std::string region = "", std::string endpoint_override = "", std::string scheme = "",
     std::string proxy_options = "", bool background_writes = true,
-    bool allow_bucket_creation = false, bool allow_bucket_deletion = false) {
+    bool allow_bucket_creation = false, bool allow_bucket_deletion = false,
+    double connect_timeout = -1, double request_timeout = -1) {
   // We need to ensure that S3 is initialized before we start messing with the
   // options
   StopIfNotOk(fs::EnsureS3Initialized());
@@ -335,6 +336,9 @@ std::shared_ptr<fs::S3FileSystem> fs___S3FileSystem__create(
   s3_opts.allow_bucket_creation = allow_bucket_creation;
   s3_opts.allow_bucket_deletion = allow_bucket_deletion;
 
+  s3_opts.request_timeout = request_timeout;
+  s3_opts.connect_timeout = connect_timeout;
+
   auto io_context = MainRThread::GetInstance().CancellableIOContext();
   return ValueOrStop(fs::S3FileSystem::Make(s3_opts, io_context));
 }
