@@ -19,4 +19,12 @@ library(testthat)
 library(arrow)
 library(tibble)
 
-test_check("arrow")
+verbose_test_output <- identical(tolower(Sys.getenv("ARROW_R_DEV", "false")), "true") ||
+  identical(tolower(Sys.getenv("ARROW_R_VERBOSE_TEST", "false")), "true")
+
+if (verbose_test_output) {
+  arrow_reporter <- MultiReporter$new(list(CheckReporter$new(), LocationReporter$new()))
+} else {
+  arrow_reporter <- check_reporter()
+}
+test_check("arrow", reporter = arrow_reporter)
