@@ -15,7 +15,6 @@
 # specific language governing permissions and limitations
 # under the License.
 
-context("Expressions")
 
 test_that("C++ expressions", {
   skip_if_not_available("dataset")
@@ -35,7 +34,7 @@ test_that("C++ expressions", {
   expect_r6_class(f == i64, "Expression")
   expect_r6_class(f == time, "Expression")
   # can't seem to make this work right now because of R Ops.method dispatch
-  # expect_r6_class(f == as.Date("2020-01-15"), "Expression")
+  # expect_r6_class(f == as.Date("2020-01-15"), "Expression") # nolint
   expect_r6_class(f == ts, "Expression")
   expect_r6_class(f <= 2L, "Expression")
   expect_r6_class(f != FALSE, "Expression")
@@ -45,25 +44,25 @@ test_that("C++ expressions", {
   expect_r6_class(!(f < 4), "Expression")
   expect_output(
     print(f > 4),
-    'Expression\n(f > 4)',
+    "Expression\n(f > 4)",
     fixed = TRUE
   )
-  expect_type_equal(
+  expect_equal(
     f$type(schema(f = float64())),
     float64()
   )
-  expect_type_equal(
+  expect_equal(
     (f > 4)$type(schema(f = float64())),
     bool()
   )
   # Interprets that as a list type
   expect_r6_class(f == c(1L, 2L), "Expression")
-  
-  expect_error(
+
+  # Non-Expression inputs are wrapped in Expression$scalar()
+  expect_equal(
     Expression$create("add", 1, 2),
-    "Expression arguments must be Expression objects"
+    Expression$create("add", Expression$scalar(1), Expression$scalar(2))
   )
-  
 })
 
 test_that("Field reference expression schemas and types", {
