@@ -31,6 +31,7 @@
 #include "arrow/c/util_internal.h"
 #include "arrow/extension_type.h"
 #include "arrow/memory_pool.h"
+#include "arrow/memory_pool_internal.h"  // for kZeroSizeArea
 #include "arrow/record_batch.h"
 #include "arrow/result.h"
 #include "arrow/stl_allocator.h"
@@ -41,6 +42,7 @@
 #include "arrow/util/logging.h"
 #include "arrow/util/macros.h"
 #include "arrow/util/small_vector.h"
+#include "arrow/util/string.h"
 #include "arrow/util/value_parsing.h"
 #include "arrow/visit_type_inline.h"
 
@@ -57,6 +59,10 @@ using internal::ArrayExportTraits;
 using internal::SchemaExportGuard;
 using internal::SchemaExportTraits;
 
+using internal::ToChars;
+
+using memory_pool::internal::kZeroSizeArea;
+
 namespace {
 
 Status ExportingNotImplemented(const DataType& type) {
@@ -334,18 +340,16 @@ struct SchemaExporter {
   Status Visit(const DoubleType& type) { return SetFormat("g"); }
 
   Status Visit(const FixedSizeBinaryType& type) {
-    return SetFormat("w:" + std::to_string(type.byte_width()));
+    return SetFormat("w:" + ToChars(type.byte_width()));
   }
 
   Status Visit(const DecimalType& type) {
     if (type.bit_width() == 128) {
       // 128 is the default bit-width
-      return SetFormat("d:" + std::to_string(type.precision()) + "," +
-                       std::to_string(type.scale()));
+      return SetFormat("d:" + ToChars(type.precision()) + "," + ToChars(type.scale()));
     } else {
-      return SetFormat("d:" + std::to_string(type.precision()) + "," +
-                       std::to_string(type.scale()) + "," +
-                       std::to_string(type.bit_width()));
+      return SetFormat("d:" + ToChars(type.precision()) + "," + ToChars(type.scale()) +
+                       "," + ToChars(type.bit_width()));
     }
   }
 
@@ -441,7 +445,7 @@ struct SchemaExporter {
   Status Visit(const LargeListType& type) { return SetFormat("+L"); }
 
   Status Visit(const FixedSizeListType& type) {
-    return SetFormat("+w:" + std::to_string(type.list_size()));
+    return SetFormat("+w:" + ToChars(type.list_size()));
   }
 
   Status Visit(const StructType& type) { return SetFormat("+s"); }
@@ -468,7 +472,7 @@ struct SchemaExporter {
       if (!first) {
         s += ",";
       }
-      s += std::to_string(code);
+      s += ToChars(code);
       first = false;
     }
     return Status::OK();
@@ -745,8 +749,9 @@ class FormatStringParser {
 
   template <typename IntType = int32_t>
   Result<std::vector<IntType>> ParseInts(std::string_view v) {
-    auto parts = Split(v);
     std::vector<IntType> result;
+    if (v.empty()) return result;
+    auto parts = Split(v);
     result.reserve(parts.size());
     for (const auto& p : parts) {
       ARROW_ASSIGN_OR_RAISE(auto i, ParseInt<IntType>(p));
@@ -1263,7 +1268,8 @@ class ImportedBuffer : public Buffer {
 };
 
 struct ArrayImporter {
-  explicit ArrayImporter(const std::shared_ptr<DataType>& type) : type_(type) {}
+  explicit ArrayImporter(const std::shared_ptr<DataType>& type)
+      : type_(type), zero_size_buffer_(std::make_shared<Buffer>(kZeroSizeArea, 0)) {}
 
   Status Import(struct ArrowArray* src) {
     if (ArrowArrayIsReleased(src)) {
@@ -1527,7 +1533,7 @@ struct ArrayImporter {
   }
 
   Status ImportNullBitmap(int32_t buffer_id = 0) {
-    RETURN_NOT_OK(ImportBitsBuffer(buffer_id));
+    RETURN_NOT_OK(ImportBitsBuffer(buffer_id, /*is_null_bitmap=*/true));
     if (data_->null_count > 0 && data_->buffers[buffer_id] == nullptr) {
       return Status::Invalid(
           "ArrowArray struct has null bitmap buffer but non-zero null_count ",
@@ -1536,15 +1542,20 @@ struct ArrayImporter {
     return Status::OK();
   }
 
-  Status ImportBitsBuffer(int32_t buffer_id) {
+  Status ImportBitsBuffer(int32_t buffer_id, bool is_null_bitmap = false) {
     // Compute visible size of buffer
-    int64_t buffer_size = bit_util::BytesForBits(c_struct_->length + c_struct_->offset);
-    return ImportBuffer(buffer_id, buffer_size);
+    int64_t buffer_size =
+        (c_struct_->length > 0)
+            ? bit_util::BytesForBits(c_struct_->length + c_struct_->offset)
+            : 0;
+    return ImportBuffer(buffer_id, buffer_size, is_null_bitmap);
   }
 
   Status ImportFixedSizeBuffer(int32_t buffer_id, int64_t byte_width) {
     // Compute visible size of buffer
-    int64_t buffer_size = byte_width * (c_struct_->length + c_struct_->offset);
+    int64_t buffer_size = (c_struct_->length > 0)
+                              ? byte_width * (c_struct_->length + c_struct_->offset)
+                              : 0;
     return ImportBuffer(buffer_id, buffer_size);
   }
 
@@ -1561,17 +1572,27 @@ struct ArrayImporter {
                                   int64_t byte_width = 1) {
     auto offsets = data_->GetValues<OffsetType>(offsets_buffer_id);
     // Compute visible size of buffer
-    int64_t buffer_size = byte_width * offsets[c_struct_->length];
+    int64_t buffer_size =
+        (c_struct_->length > 0) ? byte_width * offsets[c_struct_->length] : 0;
     return ImportBuffer(buffer_id, buffer_size);
   }
 
-  Status ImportBuffer(int32_t buffer_id, int64_t buffer_size) {
+  Status ImportBuffer(int32_t buffer_id, int64_t buffer_size,
+                      bool is_null_bitmap = false) {
     std::shared_ptr<Buffer>* out = &data_->buffers[buffer_id];
     auto data = reinterpret_cast<const uint8_t*>(c_struct_->buffers[buffer_id]);
     if (data != nullptr) {
       *out = std::make_shared<ImportedBuffer>(data, buffer_size, import_);
-    } else {
+    } else if (is_null_bitmap) {
       out->reset();
+    } else {
+      // Ensure that imported buffers are never null (except for the null bitmap)
+      if (buffer_size != 0) {
+        return Status::Invalid(
+            "ArrowArrayStruct contains null data pointer "
+            "for a buffer with non-zero computed size");
+      }
+      *out = zero_size_buffer_;
     }
     return Status::OK();
   }
@@ -1583,6 +1604,9 @@ struct ArrayImporter {
   std::shared_ptr<ImportedArrayData> import_;
   std::shared_ptr<ArrayData> data_;
   std::vector<ArrayImporter> child_importers_;
+
+  // For imported null buffer pointers
+  std::shared_ptr<Buffer> zero_size_buffer_;
 };
 
 }  // namespace
@@ -1742,7 +1766,9 @@ namespace {
 
 class ArrayStreamBatchReader : public RecordBatchReader {
  public:
-  explicit ArrayStreamBatchReader(struct ArrowArrayStream* stream) {
+  explicit ArrayStreamBatchReader(std::shared_ptr<Schema> schema,
+                                  struct ArrowArrayStream* stream)
+      : schema_(std::move(schema)) {
     ArrowArrayStreamMove(stream, &stream_);
     DCHECK(!ArrowArrayStreamIsReleased(&stream_));
   }
@@ -1754,7 +1780,7 @@ class ArrayStreamBatchReader : public RecordBatchReader {
     DCHECK(ArrowArrayStreamIsReleased(&stream_));
   }
 
-  std::shared_ptr<Schema> schema() const override { return CacheSchema(); }
+  std::shared_ptr<Schema> schema() const override { return schema_; }
 
   Status ReadNext(std::shared_ptr<RecordBatch>* batch) override {
     struct ArrowArray c_array;
@@ -1764,7 +1790,7 @@ class ArrayStreamBatchReader : public RecordBatchReader {
       batch->reset();
       return Status::OK();
     } else {
-      return ImportRecordBatch(&c_array, CacheSchema()).Value(batch);
+      return ImportRecordBatch(&c_array, schema_).Value(batch);
     }
   }
 
@@ -1775,17 +1801,30 @@ class ArrayStreamBatchReader : public RecordBatchReader {
     return Status::OK();
   }
 
- private:
-  std::shared_ptr<Schema> CacheSchema() const {
-    if (!schema_) {
-      struct ArrowSchema c_schema;
-      ARROW_CHECK_OK(StatusFromCError(stream_.get_schema(&stream_, &c_schema)));
-      schema_ = ImportSchema(&c_schema).ValueOrDie();
+  static Result<std::shared_ptr<RecordBatchReader>> Make(
+      struct ArrowArrayStream* stream) {
+    if (ArrowArrayStreamIsReleased(stream)) {
+      return Status::Invalid("Cannot import released ArrowArrayStream");
     }
-    return schema_;
+    std::shared_ptr<Schema> schema;
+    struct ArrowSchema c_schema = {};
+    auto status = StatusFromCError(stream, stream->get_schema(stream, &c_schema));
+    if (status.ok()) {
+      status = ImportSchema(&c_schema).Value(&schema);
+    }
+    if (!status.ok()) {
+      ArrowArrayStreamRelease(stream);
+      return status;
+    }
+    return std::make_shared<ArrayStreamBatchReader>(std::move(schema), stream);
   }
 
+ private:
   Status StatusFromCError(int errno_like) const {
+    return StatusFromCError(&stream_, errno_like);
+  }
+
+  static Status StatusFromCError(struct ArrowArrayStream* stream, int errno_like) {
     if (ARROW_PREDICT_TRUE(errno_like == 0)) {
       return Status::OK();
     }
@@ -1805,23 +1844,19 @@ class ArrayStreamBatchReader : public RecordBatchReader {
         code = StatusCode::IOError;
         break;
     }
-    const char* last_error = stream_.get_last_error(&stream_);
-    return Status(code, last_error ? std::string(last_error) : "");
+    const char* last_error = stream->get_last_error(stream);
+    return {code, last_error ? std::string(last_error) : ""};
   }
 
   mutable struct ArrowArrayStream stream_;
-  mutable std::shared_ptr<Schema> schema_;
+  std::shared_ptr<Schema> schema_;
 };
 
 }  // namespace
 
 Result<std::shared_ptr<RecordBatchReader>> ImportRecordBatchReader(
     struct ArrowArrayStream* stream) {
-  if (ArrowArrayStreamIsReleased(stream)) {
-    return Status::Invalid("Cannot import released ArrowArrayStream");
-  }
-  // XXX should we call get_schema() here to avoid crashing on error?
-  return std::make_shared<ArrayStreamBatchReader>(stream);
+  return ArrayStreamBatchReader::Make(stream);
 }
 
 }  // namespace arrow
