@@ -15,7 +15,10 @@
 // specific language governing permissions and limitations
 // under the License.
 
+#include "arrow/compute/exec/asof_join_node.h"
+
 #include <condition_variable>
+#include <limits>
 #include <memory>
 #include <mutex>
 #include <optional>
@@ -29,6 +32,7 @@
 #include "arrow/compute/exec/exec_plan.h"
 #include "arrow/compute/exec/key_hash.h"
 #include "arrow/compute/exec/options.h"
+#include "arrow/compute/exec/query_context.h"
 #include "arrow/compute/exec/schema_util.h"
 #include "arrow/compute/exec/util.h"
 #include "arrow/compute/light_array.h"
@@ -36,10 +40,15 @@
 #include "arrow/result.h"
 #include "arrow/status.h"
 #include "arrow/type_traits.h"
+#include "arrow/util/bit_util.h"
 #include "arrow/util/checked_cast.h"
 #include "arrow/util/future.h"
+#include "arrow/util/string.h"
 
 namespace arrow {
+
+using internal::ToChars;
+
 namespace compute {
 
 template <typename T, typename V = typename T::value_type>
@@ -56,6 +65,37 @@ typedef uint64_t ByType;
 typedef uint64_t OnType;
 typedef uint64_t HashType;
 
+/// A tolerance type with overflow-avoiding operations
+struct TolType {
+  constexpr static OnType kMinValue = std::numeric_limits<OnType>::lowest();
+  constexpr static OnType kMaxValue = std::numeric_limits<OnType>::max();
+
+  explicit TolType(int64_t tol)
+      : value(static_cast<uint64_t>(tol < 0 ? -tol : tol)), negative(tol < 0) {}
+
+  OnType value;
+  bool negative;
+
+  // an entry with a time below this threshold expires
+  inline OnType Expiry(OnType left_value) {
+    return negative ? left_value
+                    : (left_value < kMinValue + value ? kMinValue : left_value - value);
+  }
+
+  // an entry with a time after this threshold is distant
+  inline OnType Horizon(OnType left_value) {
+    return negative ? (left_value > kMaxValue - value ? kMaxValue : left_value + value)
+                    : left_value;
+  }
+
+  // true when the tolerance accepts the RHS time given the LHS one
+  inline bool Accepts(OnType left_value, OnType right_value) {
+    return negative
+               ? (left_value > right_value ? false : right_value - left_value <= value)
+               : (left_value < right_value ? false : left_value - right_value <= value);
+  }
+};
+
 // Maximum number of tables that can be joined
 #define MAX_JOIN_TABLES 64
 typedef uint64_t row_index_t;
@@ -132,6 +172,17 @@ struct MemoStore {
   // Stores last known values for all the keys
 
   struct Entry {
+    Entry() = default;
+
+    Entry(OnType time, std::shared_ptr<arrow::RecordBatch> batch, row_index_t row)
+        : time(time), batch(batch), row(row) {}
+
+    void swap(Entry& other) {
+      std::swap(time, other.time);
+      std::swap(batch, other.batch);
+      std::swap(row, other.row);
+    }
+
     // Timestamp associated with the entry
     OnType time;
 
@@ -143,31 +194,103 @@ struct MemoStore {
     row_index_t row;
   };
 
-  std::unordered_map<ByType, Entry> entries_;
+  explicit MemoStore(bool no_future)
+      : no_future_(no_future), current_time_(std::numeric_limits<OnType>::lowest()) {}
 
-  void Store(const std::shared_ptr<RecordBatch>& batch, row_index_t row, OnType time,
-             ByType key) {
-    auto& e = entries_[key];
-    // that we can do this assignment optionally, is why we
-    // can get array with using shared_ptr above (the batch
-    // shouldn't change that often)
-    if (e.batch != batch) e.batch = batch;
-    e.row = row;
-    e.time = time;
+  // true when there are no future entries, which is the case for the LHS table and the
+  // case for when the tolerance is positive. A regular non-negative-tolerance as-of-join
+  // operation requires memorizing only the most recently observed entry per key. OTOH, a
+  // negative-tolerance (future) as-of-join operation requires memorizing per-key queues
+  // of entries up to the tolerance's horizon and in particular distinguishes between the
+  // current (front-of-queue) and latest (back-of-queue) entries per key.
+  bool no_future_;
+  // the time of the current entry, defaulting to 0.
+  // when entries with a time less than T are removed, the current time is updated to the
+  // time of the next (by-time) and now-current entry or to T if no such entry exists.
+  OnType current_time_;
+  // current entry per key
+  std::unordered_map<ByType, Entry> entries_;
+  // future entries per key
+  std::unordered_map<ByType, std::queue<Entry>> future_entries_;
+  // current and future (distinct) times of existing entries
+  std::deque<OnType> times_;
+
+  void swap(MemoStore& memo) {
+    std::swap(no_future_, memo.no_future_);
+    std::swap(current_time_, memo.current_time_);
+    entries_.swap(memo.entries_);
+    future_entries_.swap(memo.future_entries_);
+    times_.swap(memo.times_);
+  }
+
+  void Store(OnType for_time, const std::shared_ptr<RecordBatch>& batch, row_index_t row,
+             OnType time, ByType key) {
+    if (no_future_ || entries_.count(key) == 0) {
+      auto& e = entries_[key];
+      // that we can do this assignment optionally, is why we
+      // can get away with using shared_ptr above (the batch
+      // shouldn't change that often)
+      if (e.batch != batch) e.batch = batch;
+      e.row = row;
+      e.time = time;
+    } else {
+      future_entries_[key].emplace(time, batch, row);
+    }
+    if (!no_future_ || times_.empty() || times_.front() != time) {
+      times_.push_back(time);
+    } else {
+      times_.front() = time;
+    }
   }
 
   std::optional<const Entry*> GetEntryForKey(ByType key) const {
     auto e = entries_.find(key);
-    if (entries_.end() == e) return std::nullopt;
-    return std::optional<const Entry*>(&e->second);
+    return entries_.end() == e ? std::nullopt : std::optional<const Entry*>(&e->second);
   }
 
-  void RemoveEntriesWithLesserTime(OnType ts) {
-    for (auto e = entries_.begin(); e != entries_.end();)
-      if (e->second.time < ts)
-        e = entries_.erase(e);
-      else
+  bool RemoveEntriesWithLesserTime(OnType ts) {
+    for (auto fe = future_entries_.begin(); fe != future_entries_.end();) {
+      auto& queue = fe->second;
+      while (!queue.empty() && queue.front().time < ts) queue.pop();
+      if (queue.empty()) {
+        fe = future_entries_.erase(fe);
+      } else {
+        ++fe;
+      }
+    }
+    for (auto e = entries_.begin(); e != entries_.end();) {
+      if (e->second.time < ts) {
+        auto fe = future_entries_.find(e->first);
+        if (fe != future_entries_.end() && !fe->second.empty()) {
+          auto& queue = fe->second;
+          e->second.swap(queue.front());
+          queue.pop();
+          ++e;
+        } else {
+          e = entries_.erase(e);
+        }
+      } else {
         ++e;
+      }
+    }
+    bool updated = false;
+    while (!times_.empty() && times_.front() < ts) {
+      current_time_ = times_.front();
+      times_.pop_front();
+      updated = true;
+    }
+    for (auto times_it = times_.begin(); times_it != times_.end(); times_it++) {
+      if (current_time_ < *times_it) {
+        current_time_ = *times_it;
+        updated = true;
+      }
+      if (*times_it > ts) break;
+    }
+    if (current_time_ < ts) {
+      current_time_ = ts;
+      updated = true;
+    }
+    return updated;
   }
 };
 
@@ -237,8 +360,8 @@ class InputState {
   // turned into output record batches.
 
  public:
-  InputState(bool must_hash, bool may_rehash, KeyHasher* key_hasher,
-             const std::shared_ptr<arrow::Schema>& schema,
+  InputState(size_t index, TolType tolerance, bool must_hash, bool may_rehash,
+             KeyHasher* key_hasher, const std::shared_ptr<arrow::Schema>& schema,
              const col_index_t time_col_index,
              const std::vector<col_index_t>& key_col_index)
       : queue_(),
@@ -249,7 +372,9 @@ class InputState {
         key_type_id_(key_col_index.size()),
         key_hasher_(key_hasher),
         must_hash_(must_hash),
-        may_rehash_(may_rehash) {
+        may_rehash_(may_rehash),
+        tolerance_(tolerance),
+        memo_(/*no_future=*/index == 0 || !tolerance.negative) {
     for (size_t k = 0; k < key_col_index_.size(); k++) {
       key_type_id_[k] = schema_->fields()[key_col_index_[k]]->type()->id();
     }
@@ -282,6 +407,20 @@ class InputState {
     return queue_.Empty();
   }
 
+  // true when the queue is empty and, when memo may have future entries (the case of a
+  // negative tolerance), when the memo is empty.
+  // used when checking whether RHS is up to date with LHS.
+  bool CurrentEmpty() const {
+    return memo_.no_future_ ? Empty() : memo_.times_.empty() && Empty();
+  }
+
+  // in case memo may not have future entries (the case of a non-negative tolerance),
+  // returns the latest time (which is current); otherwise, returns the current time.
+  // used when checking whether RHS is up to date with LHS.
+  OnType GetCurrentTime() const {
+    return memo_.no_future_ ? GetLatestTime() : memo_.current_time_;
+  }
+
   int total_batches() const { return total_batches_; }
 
   // Gets latest batch (precondition: must not be empty)
@@ -297,10 +436,10 @@ class InputState {
   }
 
   inline ByType GetLatestKey() const {
-    return GetLatestKey(queue_.UnsyncFront().get(), latest_ref_row_);
+    return GetKey(GetLatestBatch().get(), latest_ref_row_);
   }
 
-  inline ByType GetLatestKey(const RecordBatch* batch, row_index_t row) const {
+  inline ByType GetKey(const RecordBatch* batch, row_index_t row) const {
     if (must_hash_) {
       return key_hasher_->HashesFor(batch)[row];
     }
@@ -329,10 +468,10 @@ class InputState {
   }
 
   inline OnType GetLatestTime() const {
-    return GetLatestTime(queue_.UnsyncFront().get(), latest_ref_row_);
+    return GetTime(GetLatestBatch().get(), latest_ref_row_);
   }
 
-  inline ByType GetLatestTime(const RecordBatch* batch, row_index_t row) const {
+  inline ByType GetTime(const RecordBatch* batch, row_index_t row) const {
     auto data = batch->column_data(time_col_index_);
     switch (time_type_id_) {
       LATEST_VAL_CASE(INT8, time_value)
@@ -383,16 +522,18 @@ class InputState {
     return have_active_batch;
   }
 
-  // Advance the data to be immediately past the specified timestamp, update
-  // latest_time and latest_ref_row to the value that immediately pass the
-  // specified timestamp.
+  // Advance the data to be immediately past the tolerance's horizon for the specified
+  // timestamp, update latest_time and latest_ref_row to the value that immediately pass
+  // the horizon. Update the memo-store with any entries or future entries so observed.
   // Returns true if updates were made, false if not.
   Result<bool> AdvanceAndMemoize(OnType ts) {
     // Advance the right side row index until we reach the latest right row (for each key)
     // for the given left timestamp.
 
     // Check if already updated for TS (or if there is no latest)
-    if (Empty()) return false;  // can't advance if empty
+    if (Empty()) {  // can't advance if empty and no future entries
+      return memo_.no_future_ ? false : memo_.RemoveEntriesWithLesserTime(ts);
+    }
 
     // Not updated.  Try to update and possibly advance.
     bool advanced, updated = false;
@@ -402,8 +543,8 @@ class InputState {
       // Keep advancing right table until we hit the latest row that has
       // timestamp <= ts. This is because we only need the latest row for the
       // match given a left ts.
-      if (latest_time > ts) {
-        break;  // hit a future timestamp -- done updating for now
+      if (latest_time > tolerance_.Horizon(ts)) {              // hit a distant timestamp
+        if (memo_.no_future_ || !memo_.times_.empty()) break;  // no future entries
       }
       auto rb = GetLatestBatch();
       if (may_rehash_ && rb->column_data(key_col_index_[0])->GetNullCount() > 0) {
@@ -411,20 +552,30 @@ class InputState {
         may_rehash_ = false;
         Rehash();
       }
-      memo_.Store(rb, latest_ref_row_, latest_time, GetLatestKey());
-      updated = true;
+      memo_.Store(ts, rb, latest_ref_row_, latest_time, GetLatestKey());
+      updated = memo_.no_future_;
       ARROW_ASSIGN_OR_RAISE(advanced, Advance());
     } while (advanced);
+    if (!memo_.no_future_) {  // "updated" was not modified in the loop; set it here
+      updated = memo_.RemoveEntriesWithLesserTime(ts);
+    }
     return updated;
   }
 
   void Rehash() {
-    MemoStore new_memo;
-    for (const auto& entry : memo_.entries_) {
-      const auto& e = entry.second;
-      new_memo.Store(e.batch, e.row, e.time, GetLatestKey(e.batch.get(), e.row));
+    MemoStore new_memo(memo_.no_future_);
+    new_memo.current_time_ = memo_.current_time_;
+    for (auto e = memo_.entries_.begin(); e != memo_.entries_.end(); ++e) {
+      auto& entry = e->second;
+      auto new_key = GetKey(entry.batch.get(), entry.row);
+      new_memo.entries_[new_key].swap(entry);
+      auto fe = memo_.future_entries_.find(e->first);
+      if (fe != memo_.future_entries_.end()) {
+        new_memo.future_entries_[new_key].swap(fe->second);
+      }
     }
-    memo_ = new_memo;
+    memo_.times_.swap(new_memo.times_);
+    memo_.swap(new_memo);
   }
 
   Status Push(const std::shared_ptr<arrow::RecordBatch>& rb) {
@@ -484,6 +635,8 @@ class InputState {
   bool must_hash_;
   // True if by-key values may be rehashed
   bool may_rehash_;
+  // Tolerance
+  TolType tolerance_;
   // Index of the latest row reference within; if >0 then queue_ cannot be empty
   // Must be < queue_.front()->num_rows() if queue_ is non-empty
   row_index_t latest_ref_row_ = 0;
@@ -527,7 +680,7 @@ class CompositeReferenceTable {
   // Adds the latest row from the input state as a new composite reference row
   // - LHS must have a valid key,timestep,and latest rows
   // - RHS must have valid data memo'ed for the key
-  void Emplace(std::vector<std::unique_ptr<InputState>>& in, OnType tolerance) {
+  void Emplace(std::vector<std::unique_ptr<InputState>>& in, TolType tolerance) {
     DCHECK_EQ(in.size(), n_tables_);
 
     // Get the LHS key
@@ -558,7 +711,7 @@ class CompositeReferenceTable {
       std::optional<const MemoStore::Entry*> opt_entry = in[i]->GetMemoEntryForKey(key);
       if (opt_entry.has_value()) {
         DCHECK(*opt_entry);
-        if ((*opt_entry)->time + tolerance >= lhs_latest_time) {
+        if (tolerance.Accepts(lhs_latest_time, (*opt_entry)->time)) {
           // Have a valid entry
           const MemoStore::Entry* entry = *opt_entry;
           row.refs[i].batch = entry->batch.get();
@@ -608,6 +761,7 @@ class CompositeReferenceTable {
   }
 
           switch (field_type->id()) {
+            ASOFJOIN_MATERIALIZE_CASE(BOOL)
             ASOFJOIN_MATERIALIZE_CASE(INT8)
             ASOFJOIN_MATERIALIZE_CASE(INT16)
             ASOFJOIN_MATERIALIZE_CASE(INT32)
@@ -662,7 +816,6 @@ class CompositeReferenceTable {
   void AddRecordBatchRef(const std::shared_ptr<RecordBatch>& ref) {
     if (!_ptr2ref.count((uintptr_t)ref.get())) _ptr2ref[(uintptr_t)ref.get()] = ref;
   }
-
   template <class Type, class Builder = typename TypeTraits<Type>::BuilderType>
   enable_if_fixed_width_type<Type, Status> static BuilderAppend(
       Builder& builder, const std::shared_ptr<ArrayData>& source, row_index_t row) {
@@ -670,8 +823,13 @@ class CompositeReferenceTable {
       builder.UnsafeAppendNull();
       return Status::OK();
     }
-    using CType = typename TypeTraits<Type>::CType;
-    builder.UnsafeAppend(source->template GetValues<CType>(1)[row]);
+
+    if constexpr (is_boolean_type<Type>::value) {
+      builder.UnsafeAppend(bit_util::GetBit(source->template GetValues<uint8_t>(1), row));
+    } else {
+      using CType = typename TypeTraits<Type>::CType;
+      builder.UnsafeAppend(source->template GetValues<CType>(1)[row]);
+    }
     return Status::OK();
   }
 
@@ -739,9 +897,9 @@ class AsofJoinNode : public ExecNode {
       auto& rhs = *state_[i];
       if (!rhs.Finished()) {
         // If RHS is finished, then we know it's up to date
-        if (rhs.Empty())
+        if (rhs.CurrentEmpty())
           return false;  // RHS isn't finished, but is empty --> not up to date
-        if (lhs_ts >= rhs.GetLatestTime())
+        if (lhs_ts >= rhs.GetCurrentTime())
           return false;  // RHS isn't up to date (and not finished)
       }
     }
@@ -781,8 +939,9 @@ class AsofJoinNode : public ExecNode {
     // Prune memo entries that have expired (to bound memory consumption)
     if (!lhs.Empty()) {
       for (size_t i = 1; i < state_.size(); ++i) {
-        if (lhs.GetLatestTime() > tolerance_) {
-          state_[i]->RemoveMemoEntriesWithLesserTime(lhs.GetLatestTime() - tolerance_);
+        OnType ts = tolerance_.Expiry(lhs.GetLatestTime());
+        if (ts != TolType::kMinValue) {
+          state_[i]->RemoveMemoEntriesWithLesserTime(ts);
         }
       }
     }
@@ -791,15 +950,41 @@ class AsofJoinNode : public ExecNode {
     if (dst.empty()) {
       return NULLPTR;
     } else {
-      return dst.Materialize(plan()->exec_context()->memory_pool(), output_schema(),
+      return dst.Materialize(plan()->query_context()->memory_pool(), output_schema(),
                              state_);
     }
   }
 
-  void Process() {
+  template <typename Callable>
+  struct Defer {
+    Callable callable;
+    explicit Defer(Callable callable) : callable(std::move(callable)) {}
+    ~Defer() noexcept { callable(); }
+  };
+
+  void EndFromProcessThread() {
+    // We must spawn a new task to transfer off the process thread when
+    // marking this finished.  Otherwise there is a chance that doing so could
+    // mark the plan finished which may destroy the plan which will destroy this
+    // node which will cause us to join on ourselves.
+    ErrorIfNotOk(plan_->query_context()->executor()->Spawn([this] {
+      Defer cleanup([this]() { finished_.MarkFinished(); });
+      outputs_[0]->InputFinished(this, batches_produced_);
+    }));
+  }
+
+  bool CheckEnded() {
+    if (state_.at(0)->Finished()) {
+      EndFromProcessThread();
+      return false;
+    }
+    return true;
+  }
+
+  bool Process() {
     std::lock_guard<std::mutex> guard(gate_);
-    if (finished_.is_finished()) {
-      return;
+    if (!CheckEnded()) {
+      return false;
     }
 
     // Process batches while we have data
@@ -814,7 +999,8 @@ class AsofJoinNode : public ExecNode {
         outputs_[0]->InputReceived(this, std::move(out_b));
       } else {
         ErrorIfNotOk(result.status());
-        return;
+        EndFromProcessThread();
+        return false;
       }
     }
 
@@ -823,18 +1009,24 @@ class AsofJoinNode : public ExecNode {
     //
     // It may happen here in cases where InputFinished was called before we were finished
     // producing results (so we didn't know the output size at that time)
-    if (state_.at(0)->Finished()) {
-      outputs_[0]->InputFinished(this, batches_produced_);
-      finished_.MarkFinished();
+    if (!CheckEnded()) {
+      return false;
     }
+
+    // There is no more we can do now but there is still work remaining for later when
+    // more data arrives.
+    return true;
   }
 
   void ProcessThread() {
     for (;;) {
       if (!process_.Pop()) {
+        EndFromProcessThread();
+        return;
+      }
+      if (!Process()) {
         return;
       }
-      Process();
     }
   }
 
@@ -844,17 +1036,18 @@ class AsofJoinNode : public ExecNode {
   AsofJoinNode(ExecPlan* plan, NodeVector inputs, std::vector<std::string> input_labels,
                const std::vector<col_index_t>& indices_of_on_key,
                const std::vector<std::vector<col_index_t>>& indices_of_by_key,
-               OnType tolerance, std::shared_ptr<Schema> output_schema,
+               TolType tolerance, std::shared_ptr<Schema> output_schema,
                std::vector<std::unique_ptr<KeyHasher>> key_hashers, bool must_hash,
                bool may_rehash);
 
   Status Init() override {
     auto inputs = this->inputs();
     for (size_t i = 0; i < inputs.size(); i++) {
-      RETURN_NOT_OK(key_hashers_[i]->Init(plan()->exec_context(), output_schema()));
+      RETURN_NOT_OK(key_hashers_[i]->Init(plan()->query_context()->exec_context(),
+                                          output_schema()));
       state_.push_back(std::make_unique<InputState>(
-          must_hash_, may_rehash_, key_hashers_[i].get(), inputs[i]->output_schema(),
-          indices_of_on_key_[i], indices_of_by_key_[i]));
+          i, tolerance_, must_hash_, may_rehash_, key_hashers_[i].get(),
+          inputs[i]->output_schema(), indices_of_on_key_[i], indices_of_by_key_[i]));
     }
 
     col_index_t dst_offset = 0;
@@ -924,6 +1117,7 @@ class AsofJoinNode : public ExecNode {
 
   static Status is_valid_data_field(const std::shared_ptr<Field>& field) {
     switch (field->type()->id()) {
+      case Type::BOOL:
       case Type::INT8:
       case Type::INT16:
       case Type::INT32:
@@ -951,17 +1145,16 @@ class AsofJoinNode : public ExecNode {
   }
 
   static arrow::Result<std::shared_ptr<Schema>> MakeOutputSchema(
-      const std::vector<ExecNode*>& inputs,
+      const std::vector<std::shared_ptr<Schema>> input_schema,
       const std::vector<col_index_t>& indices_of_on_key,
       const std::vector<std::vector<col_index_t>>& indices_of_by_key) {
     std::vector<std::shared_ptr<arrow::Field>> fields;
 
-    size_t n_by = indices_of_by_key[0].size();
+    size_t n_by = indices_of_by_key.size() == 0 ? 0 : indices_of_by_key[0].size();
     const DataType* on_key_type = NULLPTR;
     std::vector<const DataType*> by_key_type(n_by, NULLPTR);
     // Take all non-key, non-time RHS fields
-    for (size_t j = 0; j < inputs.size(); ++j) {
-      const auto& input_schema = inputs[j]->output_schema();
+    for (size_t j = 0; j < input_schema.size(); ++j) {
       const auto& on_field_ix = indices_of_on_key[j];
       const auto& by_field_ix = indices_of_by_key[j];
 
@@ -969,10 +1162,10 @@ class AsofJoinNode : public ExecNode {
         return Status::Invalid("Missing join key on table ", j);
       }
 
-      const auto& on_field = input_schema->fields()[on_field_ix];
+      const auto& on_field = input_schema[j]->fields()[on_field_ix];
       std::vector<const Field*> by_field(n_by);
       for (size_t k = 0; k < n_by; k++) {
-        by_field[k] = input_schema->fields()[by_field_ix[k]].get();
+        by_field[k] = input_schema[j]->fields()[by_field_ix[k]].get();
       }
 
       if (on_key_type == NULLPTR) {
@@ -992,8 +1185,8 @@ class AsofJoinNode : public ExecNode {
         }
       }
 
-      for (int i = 0; i < input_schema->num_fields(); ++i) {
-        const auto field = input_schema->field(i);
+      for (int i = 0; i < input_schema[j]->num_fields(); ++i) {
+        const auto field = input_schema[j]->field(i);
         if (i == on_field_ix) {
           ARROW_RETURN_NOT_OK(is_valid_on_field(field));
           // Only add on field from the left table
@@ -1030,34 +1223,75 @@ class AsofJoinNode : public ExecNode {
     return match.indices()[0];
   }
 
-  static arrow::Result<ExecNode*> Make(ExecPlan* plan, std::vector<ExecNode*> inputs,
-                                       const ExecNodeOptions& options) {
-    DCHECK_GE(inputs.size(), 2) << "Must have at least two inputs";
-
-    const auto& join_options = checked_cast<const AsofJoinNodeOptions&>(options);
-    if (join_options.tolerance < 0) {
-      return Status::Invalid("AsOfJoin tolerance must be non-negative but is ",
-                             join_options.tolerance);
+  static Result<size_t> GetByKeySize(
+      const std::vector<asofjoin::AsofJoinKeys>& input_keys) {
+    size_t n_by = 0;
+    for (size_t i = 0; i < input_keys.size(); ++i) {
+      const auto& by_key = input_keys[i].by_key;
+      if (i == 0) {
+        n_by = by_key.size();
+      } else if (n_by != by_key.size()) {
+        return Status::Invalid("inconsistent size of by-key across inputs");
+      }
     }
+    return n_by;
+  }
 
-    size_t n_input = inputs.size(), n_by = join_options.by_key.size();
-    std::vector<std::string> input_labels(n_input);
+  static Result<std::vector<col_index_t>> GetIndicesOfOnKey(
+      const std::vector<std::shared_ptr<Schema>>& input_schema,
+      const std::vector<asofjoin::AsofJoinKeys>& input_keys) {
+    if (input_schema.size() != input_keys.size()) {
+      return Status::Invalid("mismatching number of input schema and keys");
+    }
+    size_t n_input = input_schema.size();
     std::vector<col_index_t> indices_of_on_key(n_input);
+    for (size_t i = 0; i < n_input; ++i) {
+      const auto& on_key = input_keys[i].on_key;
+      ARROW_ASSIGN_OR_RAISE(indices_of_on_key[i],
+                            FindColIndex(*input_schema[i], on_key, "on"));
+    }
+    return indices_of_on_key;
+  }
+
+  static Result<std::vector<std::vector<col_index_t>>> GetIndicesOfByKey(
+      const std::vector<std::shared_ptr<Schema>>& input_schema,
+      const std::vector<asofjoin::AsofJoinKeys>& input_keys) {
+    if (input_schema.size() != input_keys.size()) {
+      return Status::Invalid("mismatching number of input schema and keys");
+    }
+    ARROW_ASSIGN_OR_RAISE(size_t n_by, GetByKeySize(input_keys));
+    size_t n_input = input_schema.size();
     std::vector<std::vector<col_index_t>> indices_of_by_key(
         n_input, std::vector<col_index_t>(n_by));
     for (size_t i = 0; i < n_input; ++i) {
-      input_labels[i] = i == 0 ? "left" : "right_" + std::to_string(i);
-      const Schema& input_schema = *inputs[i]->output_schema();
-      ARROW_ASSIGN_OR_RAISE(indices_of_on_key[i],
-                            FindColIndex(input_schema, join_options.on_key, "on"));
       for (size_t k = 0; k < n_by; k++) {
+        const auto& by_key = input_keys[i].by_key;
         ARROW_ASSIGN_OR_RAISE(indices_of_by_key[i][k],
-                              FindColIndex(input_schema, join_options.by_key[k], "by"));
+                              FindColIndex(*input_schema[i], by_key[k], "by"));
       }
     }
+    return indices_of_by_key;
+  }
 
-    ARROW_ASSIGN_OR_RAISE(std::shared_ptr<Schema> output_schema,
-                          MakeOutputSchema(inputs, indices_of_on_key, indices_of_by_key));
+  static arrow::Result<ExecNode*> Make(ExecPlan* plan, std::vector<ExecNode*> inputs,
+                                       const ExecNodeOptions& options) {
+    DCHECK_GE(inputs.size(), 2) << "Must have at least two inputs";
+    const auto& join_options = checked_cast<const AsofJoinNodeOptions&>(options);
+    ARROW_ASSIGN_OR_RAISE(size_t n_by, GetByKeySize(join_options.input_keys));
+    size_t n_input = inputs.size();
+    std::vector<std::string> input_labels(n_input);
+    std::vector<std::shared_ptr<Schema>> input_schema(n_input);
+    for (size_t i = 0; i < n_input; ++i) {
+      input_labels[i] = i == 0 ? "left" : "right_" + ToChars(i);
+      input_schema[i] = inputs[i]->output_schema();
+    }
+    ARROW_ASSIGN_OR_RAISE(std::vector<col_index_t> indices_of_on_key,
+                          GetIndicesOfOnKey(input_schema, join_options.input_keys));
+    ARROW_ASSIGN_OR_RAISE(std::vector<std::vector<col_index_t>> indices_of_by_key,
+                          GetIndicesOfByKey(input_schema, join_options.input_keys));
+    ARROW_ASSIGN_OR_RAISE(
+        std::shared_ptr<Schema> output_schema,
+        MakeOutputSchema(input_schema, indices_of_on_key, indices_of_by_key));
 
     std::vector<std::unique_ptr<KeyHasher>> key_hashers;
     for (size_t i = 0; i < n_input; i++) {
@@ -1071,7 +1305,7 @@ class AsofJoinNode : public ExecNode {
     bool may_rehash = n_by == 1 && !must_hash;
     return plan->EmplaceNode<AsofJoinNode>(
         plan, inputs, std::move(input_labels), std::move(indices_of_on_key),
-        std::move(indices_of_by_key), time_value(join_options.tolerance),
+        std::move(indices_of_by_key), TolType(join_options.tolerance),
         std::move(output_schema), std::move(key_hashers), must_hash, may_rehash);
   }
 
@@ -1107,10 +1341,7 @@ class AsofJoinNode : public ExecNode {
     // finished.
     process_.Push(true);
   }
-  Status StartProducing() override {
-    finished_ = arrow::Future<>::Make();
-    return Status::OK();
-  }
+  Status StartProducing() override { return Status::OK(); }
   void PauseProducing(ExecNode* output, int32_t counter) override {}
   void ResumeProducing(ExecNode* output, int32_t counter) override {}
   void StopProducing(ExecNode* output) override {
@@ -1124,7 +1355,6 @@ class AsofJoinNode : public ExecNode {
   arrow::Future<> finished() override { return finished_; }
 
  private:
-  arrow::Future<> finished_;
   std::vector<col_index_t> indices_of_on_key_;
   std::vector<std::vector<col_index_t>> indices_of_by_key_;
   std::vector<std::unique_ptr<KeyHasher>> key_hashers_;
@@ -1134,7 +1364,7 @@ class AsofJoinNode : public ExecNode {
   // Each input state correponds to an input table
   std::vector<std::unique_ptr<InputState>> state_;
   std::mutex gate_;
-  OnType tolerance_;
+  TolType tolerance_;
 
   // Queue for triggering processing of a given input
   // (a false value is a poison pill)
@@ -1150,7 +1380,7 @@ AsofJoinNode::AsofJoinNode(ExecPlan* plan, NodeVector inputs,
                            std::vector<std::string> input_labels,
                            const std::vector<col_index_t>& indices_of_on_key,
                            const std::vector<std::vector<col_index_t>>& indices_of_by_key,
-                           OnType tolerance, std::shared_ptr<Schema> output_schema,
+                           TolType tolerance, std::shared_ptr<Schema> output_schema,
                            std::vector<std::unique_ptr<KeyHasher>> key_hashers,
                            bool must_hash, bool may_rehash)
     : ExecNode(plan, inputs, input_labels,
@@ -1163,9 +1393,7 @@ AsofJoinNode::AsofJoinNode(ExecPlan* plan, NodeVector inputs,
       may_rehash_(may_rehash),
       tolerance_(tolerance),
       process_(),
-      process_thread_(&AsofJoinNode::ProcessThreadWrapper, this) {
-  finished_ = arrow::Future<>::MakeFinished();
-}
+      process_thread_(&AsofJoinNode::ProcessThreadWrapper, this) {}
 
 namespace internal {
 void RegisterAsofJoinNode(ExecFactoryRegistry* registry) {
@@ -1173,5 +1401,20 @@ void RegisterAsofJoinNode(ExecFactoryRegistry* registry) {
 }
 }  // namespace internal
 
+namespace asofjoin {
+
+Result<std::shared_ptr<Schema>> MakeOutputSchema(
+    const std::vector<std::shared_ptr<Schema>>& input_schema,
+    const std::vector<AsofJoinKeys>& input_keys) {
+  ARROW_ASSIGN_OR_RAISE(std::vector<col_index_t> indices_of_on_key,
+                        AsofJoinNode::GetIndicesOfOnKey(input_schema, input_keys));
+  ARROW_ASSIGN_OR_RAISE(std::vector<std::vector<col_index_t>> indices_of_by_key,
+                        AsofJoinNode::GetIndicesOfByKey(input_schema, input_keys));
+  return AsofJoinNode::MakeOutputSchema(input_schema, indices_of_on_key,
+                                        indices_of_by_key);
+}
+
+}  // namespace asofjoin
+
 }  // namespace compute
 }  // namespace arrow
