@@ -2022,7 +2022,7 @@ Status JoinProbeProcessor::OnFinished() {
 
 class SwissJoin : public HashJoinImpl {
  public:
-  Status Init(ExecContext* ctx, JoinType join_type, size_t num_threads,
+  Status Init(QueryContext* ctx, JoinType join_type, size_t num_threads,
               const HashJoinProjectionMaps* proj_map_left,
               const HashJoinProjectionMaps* proj_map_right,
               std::vector<JoinKeyCmp> key_cmp, Expression filter,
@@ -2067,8 +2067,6 @@ class SwissJoin : public HashJoinImpl {
     for (int i = 0; i < num_threads_; ++i) {
       local_states_[i].hash_table_ready = false;
       local_states_[i].num_output_batches = 0;
-      RETURN_NOT_OK(CancelIfNotOK(local_states_[i].temp_stack.Init(
-          pool_, 1024 + 64 * util::MiniBatch::kMiniBatchLength)));
       local_states_[i].materialize.Init(pool_, proj_map_left, proj_map_right);
     }
 
@@ -2116,10 +2114,12 @@ class SwissJoin : public HashJoinImpl {
 
     ExecBatch keypayload_batch;
     ARROW_ASSIGN_OR_RAISE(keypayload_batch, KeyPayloadFromInput(/*side=*/0, &batch));
+    ARROW_ASSIGN_OR_RAISE(util::TempVectorStack * temp_stack,
+                          ctx_->GetTempStack(thread_index));
 
-    return CancelIfNotOK(probe_processor_.OnNextBatch(
-        thread_index, keypayload_batch, &local_states_[thread_index].temp_stack,
-        &local_states_[thread_index].temp_column_arrays));
+    return CancelIfNotOK(
+        probe_processor_.OnNextBatch(thread_index, keypayload_batch, temp_stack,
+                                     &local_states_[thread_index].temp_column_arrays));
   }
 
   Status ProbingFinished(size_t thread_index) override {
@@ -2225,9 +2225,11 @@ class SwissJoin : public HashJoinImpl {
             input_batch.values[schema->num_cols(HashJoinProjection::KEY) + icol];
       }
     }
+    ARROW_ASSIGN_OR_RAISE(util::TempVectorStack * temp_stack,
+                          ctx_->GetTempStack(thread_id));
     RETURN_NOT_OK(CancelIfNotOK(hash_table_build_.PushNextBatch(
         static_cast<int64_t>(thread_id), key_batch, no_payload ? nullptr : &payload_batch,
-        &local_states_[thread_id].temp_stack)));
+        temp_stack)));
 
     // Release input batch
     //
@@ -2259,7 +2261,9 @@ class SwissJoin : public HashJoinImpl {
 
   Status MergeFinished(size_t thread_id) {
     RETURN_NOT_OK(status());
-    hash_table_build_.FinishPrtnMerge(&local_states_[thread_id].temp_stack);
+    ARROW_ASSIGN_OR_RAISE(util::TempVectorStack * temp_stack,
+                          ctx_->GetTempStack(thread_id));
+    hash_table_build_.FinishPrtnMerge(temp_stack);
     return CancelIfNotOK(OnBuildHashTableFinished(static_cast<int64_t>(thread_id)));
   }
 
@@ -2311,7 +2315,8 @@ class SwissJoin : public HashJoinImpl {
         std::min((task_id + 1) * kNumRowsPerScanTask, hash_table_.num_rows());
     // Get thread index and related temp vector stack
     //
-    util::TempVectorStack* temp_stack = &local_states_[thread_id].temp_stack;
+    ARROW_ASSIGN_OR_RAISE(util::TempVectorStack * temp_stack,
+                          ctx_->GetTempStack(thread_id));
 
     // Split into mini-batches
     //
@@ -2467,7 +2472,7 @@ class SwissJoin : public HashJoinImpl {
 
   static constexpr int kNumRowsPerScanTask = 512 * 1024;
 
-  ExecContext* ctx_;
+  QueryContext* ctx_;
   int64_t hardware_flags_;
   MemoryPool* pool_;
   int num_threads_;
@@ -2489,7 +2494,6 @@ class SwissJoin : public HashJoinImpl {
 
   struct ThreadLocalState {
     JoinResultMaterialize materialize;
-    util::TempVectorStack temp_stack;
     std::vector<KeyColumnArray> temp_column_arrays;
     int64_t num_output_batches;
     bool hash_table_ready;
