@@ -67,8 +67,11 @@ struct SetLookupState : public KernelState {
     auto visit_valid = [&](T v) {
       const auto memo_size = static_cast<int32_t>(memo_index_to_value_index.size());
       int32_t unused_memo_index;
-      auto on_found = [&](int32_t memo_index) { DCHECK_LT(memo_index, memo_size); };
-      auto on_not_found = [&](int32_t memo_index) {
+      // (capture `memo_size` by value because of ARROW-17567)
+      auto on_found = [&, memo_size](int32_t memo_index) {
+        DCHECK_LT(memo_index, memo_size);
+      };
+      auto on_not_found = [&, memo_size](int32_t memo_index) {
         DCHECK_EQ(memo_index, memo_size);
         memo_index_to_value_index.push_back(index);
       };
@@ -79,8 +82,10 @@ struct SetLookupState : public KernelState {
     };
     auto visit_null = [&]() {
       const auto memo_size = static_cast<int32_t>(memo_index_to_value_index.size());
-      auto on_found = [&](int32_t memo_index) { DCHECK_LT(memo_index, memo_size); };
-      auto on_not_found = [&](int32_t memo_index) {
+      auto on_found = [&, memo_size](int32_t memo_index) {
+        DCHECK_LT(memo_index, memo_size);
+      };
+      auto on_not_found = [&, memo_size](int32_t memo_index) {
         DCHECK_EQ(memo_index, memo_size);
         memo_index_to_value_index.push_back(index);
       };
