@@ -32,9 +32,13 @@
 #include "arrow/util/bit_util.h"
 #include "arrow/util/io_util.h"
 #include "arrow/util/logging.h"
+#include "arrow/util/string.h"
 #include "arrow/util/uri.h"
 
 namespace arrow {
+
+using internal::ToChars;
+
 namespace flight {
 namespace transport {
 namespace ucx {
@@ -122,7 +126,7 @@ arrow::Result<std::string> SockaddrToString(const struct sockaddr_storage& addre
   DCHECK_NE(pos, std::string::npos);
   result[pos] = ':';
   result.resize(pos + 1);
-  result += std::to_string(port);
+  result += ToChars(port);
   return result;
 }
 
