@@ -59,7 +59,7 @@ namespace memory_pool {
 
 namespace internal {
 
-alignas(kAlignment) int64_t zero_size_area[1] = {kDebugXorSuffix};
+alignas(kDefaultBufferAlignment) int64_t zero_size_area[1] = {kDebugXorSuffix};
 
 }  // namespace internal
 
@@ -224,14 +224,14 @@ bool IsDebugEnabled() {
 template <typename WrappedAllocator>
 class DebugAllocator {
  public:
-  static Status AllocateAligned(int64_t size, uint8_t** out) {
+  static Status AllocateAligned(int64_t size, int64_t alignment, uint8_t** out) {
     if (size == 0) {
       *out = memory_pool::internal::kZeroSizeArea;
     } else {
       ARROW_ASSIGN_OR_RAISE(int64_t raw_size, RawSize(size));
       DCHECK(raw_size > size) << "bug in raw size computation: " << raw_size
                               << " for size " << size;
-      RETURN_NOT_OK(WrappedAllocator::AllocateAligned(raw_size, out));
+      RETURN_NOT_OK(WrappedAllocator::AllocateAligned(raw_size, alignment, out));
       InitAllocatedArea(*out, size);
     }
     return Status::OK();
@@ -239,31 +239,32 @@ class DebugAllocator {
 
   static void ReleaseUnused() { WrappedAllocator::ReleaseUnused(); }
 
-  static Status ReallocateAligned(int64_t old_size, int64_t new_size, uint8_t** ptr) {
+  static Status ReallocateAligned(int64_t old_size, int64_t new_size, int64_t alignment,
+                                  uint8_t** ptr) {
     CheckAllocatedArea(*ptr, old_size, "reallocation");
     if (*ptr == memory_pool::internal::kZeroSizeArea) {
-      return AllocateAligned(new_size, ptr);
+      return AllocateAligned(new_size, alignment, ptr);
     }
     if (new_size == 0) {
       // Note that an overflow check isn't needed as `old_size` is supposed to have
       // been successfully passed to AllocateAligned() before.
-      WrappedAllocator::DeallocateAligned(*ptr, old_size + kOverhead);
+      WrappedAllocator::DeallocateAligned(*ptr, old_size + kOverhead, alignment);
       *ptr = memory_pool::internal::kZeroSizeArea;
       return Status::OK();
     }
     ARROW_ASSIGN_OR_RAISE(int64_t raw_new_size, RawSize(new_size));
     DCHECK(raw_new_size > new_size)
         << "bug in raw size computation: " << raw_new_size << " for size " << new_size;
-    RETURN_NOT_OK(
-        WrappedAllocator::ReallocateAligned(old_size + kOverhead, raw_new_size, ptr));
+    RETURN_NOT_OK(WrappedAllocator::ReallocateAligned(old_size + kOverhead, raw_new_size,
+                                                      alignment, ptr));
     InitAllocatedArea(*ptr, new_size);
     return Status::OK();
   }
 
-  static void DeallocateAligned(uint8_t* ptr, int64_t size) {
+  static void DeallocateAligned(uint8_t* ptr, int64_t size, int64_t alignment) {
     CheckAllocatedArea(ptr, size, "deallocation");
     if (ptr != memory_pool::internal::kZeroSizeArea) {
-      WrappedAllocator::DeallocateAligned(ptr, size + kOverhead);
+      WrappedAllocator::DeallocateAligned(ptr, size + kOverhead, alignment);
     }
   }
 
@@ -299,7 +300,7 @@ class SystemAllocator {
  public:
   // Allocate memory according to the alignment requirements for Arrow
   // (as of May 2016 64 bytes)
-  static Status AllocateAligned(int64_t size, uint8_t** out) {
+  static Status AllocateAligned(int64_t size, int64_t alignment, uint8_t** out) {
     if (size == 0) {
       *out = memory_pool::internal::kZeroSizeArea;
       return Status::OK();
@@ -307,19 +308,19 @@ class SystemAllocator {
 #ifdef _WIN32
     // Special code path for Windows
     *out = reinterpret_cast<uint8_t*>(
-        _aligned_malloc(static_cast<size_t>(size), memory_pool::internal::kAlignment));
+        _aligned_malloc(static_cast<size_t>(size), static_cast<size_t>(alignment)));
     if (!*out) {
       return Status::OutOfMemory("malloc of size ", size, " failed");
     }
 #elif defined(sun) || defined(__sun)
     *out = reinterpret_cast<uint8_t*>(
-        memalign(memory_pool::internal::kAlignment, static_cast<size_t>(size)));
+        memalign(static_cast<size_t>(alignment), static_cast<size_t>(size)));
     if (!*out) {
       return Status::OutOfMemory("malloc of size ", size, " failed");
     }
 #else
     const int result =
-        posix_memalign(reinterpret_cast<void**>(out), memory_pool::internal::kAlignment,
+        posix_memalign(reinterpret_cast<void**>(out), static_cast<size_t>(alignment),
                        static_cast<size_t>(size));
     if (result == ENOMEM) {
       return Status::OutOfMemory("malloc of size ", size, " failed");
@@ -327,20 +328,21 @@ class SystemAllocator {
 
     if (result == EINVAL) {
       return Status::Invalid("invalid alignment parameter: ",
-                             memory_pool::internal::kAlignment);
+                             static_cast<size_t>(alignment));
     }
 #endif
     return Status::OK();
   }
 
-  static Status ReallocateAligned(int64_t old_size, int64_t new_size, uint8_t** ptr) {
+  static Status ReallocateAligned(int64_t old_size, int64_t new_size, int64_t alignment,
+                                  uint8_t** ptr) {
     uint8_t* previous_ptr = *ptr;
     if (previous_ptr == memory_pool::internal::kZeroSizeArea) {
       DCHECK_EQ(old_size, 0);
-      return AllocateAligned(new_size, ptr);
+      return AllocateAligned(new_size, alignment, ptr);
     }
     if (new_size == 0) {
-      DeallocateAligned(previous_ptr, old_size);
+      DeallocateAligned(previous_ptr, old_size, alignment);
       *ptr = memory_pool::internal::kZeroSizeArea;
       return Status::OK();
     }
@@ -348,7 +350,7 @@ class SystemAllocator {
 
     // Allocate new chunk
     uint8_t* out = nullptr;
-    RETURN_NOT_OK(AllocateAligned(new_size, &out));
+    RETURN_NOT_OK(AllocateAligned(new_size, alignment, &out));
     DCHECK(out);
     // Copy contents and release old memory chunk
     memcpy(out, *ptr, static_cast<size_t>(std::min(new_size, old_size)));
@@ -361,7 +363,7 @@ class SystemAllocator {
     return Status::OK();
   }
 
-  static void DeallocateAligned(uint8_t* ptr, int64_t size) {
+  static void DeallocateAligned(uint8_t* ptr, int64_t size, int64_t /*alignment*/) {
     if (ptr == memory_pool::internal::kZeroSizeArea) {
       DCHECK_EQ(size, 0);
     } else {
@@ -387,13 +389,13 @@ class SystemAllocator {
 // Helper class directing allocations to the mimalloc allocator.
 class MimallocAllocator {
  public:
-  static Status AllocateAligned(int64_t size, uint8_t** out) {
+  static Status AllocateAligned(int64_t size, int64_t alignment, uint8_t** out) {
     if (size == 0) {
       *out = memory_pool::internal::kZeroSizeArea;
       return Status::OK();
     }
     *out = reinterpret_cast<uint8_t*>(
-        mi_malloc_aligned(static_cast<size_t>(size), memory_pool::internal::kAlignment));
+        mi_malloc_aligned(static_cast<size_t>(size), static_cast<size_t>(alignment)));
     if (*out == NULL) {
       return Status::OutOfMemory("malloc of size ", size, " failed");
     }
@@ -402,19 +404,20 @@ class MimallocAllocator {
 
   static void ReleaseUnused() { mi_collect(true); }
 
-  static Status ReallocateAligned(int64_t old_size, int64_t new_size, uint8_t** ptr) {
+  static Status ReallocateAligned(int64_t old_size, int64_t new_size, int64_t alignment,
+                                  uint8_t** ptr) {
     uint8_t* previous_ptr = *ptr;
     if (previous_ptr == memory_pool::internal::kZeroSizeArea) {
       DCHECK_EQ(old_size, 0);
-      return AllocateAligned(new_size, ptr);
+      return AllocateAligned(new_size, alignment, ptr);
     }
     if (new_size == 0) {
-      DeallocateAligned(previous_ptr, old_size);
+      DeallocateAligned(previous_ptr, old_size, alignment);
       *ptr = memory_pool::internal::kZeroSizeArea;
       return Status::OK();
     }
-    *ptr = reinterpret_cast<uint8_t*>(mi_realloc_aligned(
-        previous_ptr, static_cast<size_t>(new_size), memory_pool::internal::kAlignment));
+    *ptr = reinterpret_cast<uint8_t*>(
+        mi_realloc_aligned(previous_ptr, static_cast<size_t>(new_size), alignment));
     if (*ptr == NULL) {
       *ptr = previous_ptr;
       return Status::OutOfMemory("realloc of size ", new_size, " failed");
@@ -422,7 +425,7 @@ class MimallocAllocator {
     return Status::OK();
   }
 
-  static void DeallocateAligned(uint8_t* ptr, int64_t size) {
+  static void DeallocateAligned(uint8_t* ptr, int64_t size, int64_t /*alignment*/) {
     if (ptr == memory_pool::internal::kZeroSizeArea) {
       DCHECK_EQ(size, 0);
     } else {
@@ -452,14 +455,14 @@ class BaseMemoryPoolImpl : public MemoryPool {
  public:
   ~BaseMemoryPoolImpl() override {}
 
-  Status Allocate(int64_t size, uint8_t** out) override {
+  Status Allocate(int64_t size, int64_t alignment, uint8_t** out) override {
     if (size < 0) {
       return Status::Invalid("negative malloc size");
     }
     if (static_cast<uint64_t>(size) >= std::numeric_limits<size_t>::max()) {
       return Status::OutOfMemory("malloc size overflows size_t");
     }
-    RETURN_NOT_OK(Allocator::AllocateAligned(size, out));
+    RETURN_NOT_OK(Allocator::AllocateAligned(size, alignment, out));
 #ifndef NDEBUG
     // Poison data
     if (size > 0) {
@@ -473,14 +476,15 @@ class BaseMemoryPoolImpl : public MemoryPool {
     return Status::OK();
   }
 
-  Status Reallocate(int64_t old_size, int64_t new_size, uint8_t** ptr) override {
+  Status Reallocate(int64_t old_size, int64_t new_size, int64_t alignment,
+                    uint8_t** ptr) override {
     if (new_size < 0) {
       return Status::Invalid("negative realloc size");
     }
     if (static_cast<uint64_t>(new_size) >= std::numeric_limits<size_t>::max()) {
       return Status::OutOfMemory("realloc overflows size_t");
     }
-    RETURN_NOT_OK(Allocator::ReallocateAligned(old_size, new_size, ptr));
+    RETURN_NOT_OK(Allocator::ReallocateAligned(old_size, new_size, alignment, ptr));
 #ifndef NDEBUG
     // Poison data
     if (new_size > old_size) {
@@ -494,7 +498,7 @@ class BaseMemoryPoolImpl : public MemoryPool {
     return Status::OK();
   }
 
-  void Free(uint8_t* buffer, int64_t size) override {
+  void Free(uint8_t* buffer, int64_t size, int64_t alignment) override {
 #ifndef NDEBUG
     // Poison data
     if (size > 0) {
@@ -503,7 +507,7 @@ class BaseMemoryPoolImpl : public MemoryPool {
       buffer[size - 1] = kDeallocPoison;
     }
 #endif
-    Allocator::DeallocateAligned(buffer, size);
+    Allocator::DeallocateAligned(buffer, size, alignment);
 
     stats_.UpdateAllocatedBytes(-size);
   }
@@ -697,22 +701,23 @@ Result<std::string> jemalloc_stats_string(const char* opts) {
 
 LoggingMemoryPool::LoggingMemoryPool(MemoryPool* pool) : pool_(pool) {}
 
-Status LoggingMemoryPool::Allocate(int64_t size, uint8_t** out) {
-  Status s = pool_->Allocate(size, out);
-  std::cout << "Allocate: size = " << size << std::endl;
+Status LoggingMemoryPool::Allocate(int64_t size, int64_t alignment, uint8_t** out) {
+  Status s = pool_->Allocate(size, alignment, out);
+  std::cout << "Allocate: size = " << size << ", alignment = " << alignment << std::endl;
   return s;
 }
 
-Status LoggingMemoryPool::Reallocate(int64_t old_size, int64_t new_size, uint8_t** ptr) {
+Status LoggingMemoryPool::Reallocate(int64_t old_size, int64_t new_size,
+                                     int64_t alignment, uint8_t** ptr) {
   Status s = pool_->Reallocate(old_size, new_size, ptr);
-  std::cout << "Reallocate: old_size = " << old_size << " - new_size = " << new_size
-            << std::endl;
+  std::cout << "Reallocate: old_size = " << old_size << ", new_size = " << new_size
+            << ", alignment = " << alignment << std::endl;
   return s;
 }
 
-void LoggingMemoryPool::Free(uint8_t* buffer, int64_t size) {
-  pool_->Free(buffer, size);
-  std::cout << "Free: size = " << size << std::endl;
+void LoggingMemoryPool::Free(uint8_t* buffer, int64_t size, int64_t alignment) {
+  pool_->Free(buffer, size, alignment);
+  std::cout << "Free: size = " << size << ", alignment = " << alignment << std::endl;
 }
 
 int64_t LoggingMemoryPool::bytes_allocated() const {
@@ -736,20 +741,21 @@ class ProxyMemoryPool::ProxyMemoryPoolImpl {
  public:
   explicit ProxyMemoryPoolImpl(MemoryPool* pool) : pool_(pool) {}
 
-  Status Allocate(int64_t size, uint8_t** out) {
-    RETURN_NOT_OK(pool_->Allocate(size, out));
+  Status Allocate(int64_t size, int64_t alignment, uint8_t** out) {
+    RETURN_NOT_OK(pool_->Allocate(size, alignment, out));
     stats_.UpdateAllocatedBytes(size);
     return Status::OK();
   }
 
-  Status Reallocate(int64_t old_size, int64_t new_size, uint8_t** ptr) {
-    RETURN_NOT_OK(pool_->Reallocate(old_size, new_size, ptr));
+  Status Reallocate(int64_t old_size, int64_t new_size, int64_t alignment,
+                    uint8_t** ptr) {
+    RETURN_NOT_OK(pool_->Reallocate(old_size, new_size, alignment, ptr));
     stats_.UpdateAllocatedBytes(new_size - old_size);
     return Status::OK();
   }
 
-  void Free(uint8_t* buffer, int64_t size) {
-    pool_->Free(buffer, size);
+  void Free(uint8_t* buffer, int64_t size, int64_t alignment) {
+    pool_->Free(buffer, size, alignment);
     stats_.UpdateAllocatedBytes(-size);
   }
 
@@ -770,16 +776,17 @@ ProxyMemoryPool::ProxyMemoryPool(MemoryPool* pool) {
 
 ProxyMemoryPool::~ProxyMemoryPool() {}
 
-Status ProxyMemoryPool::Allocate(int64_t size, uint8_t** out) {
-  return impl_->Allocate(size, out);
+Status ProxyMemoryPool::Allocate(int64_t size, int64_t alignment, uint8_t** out) {
+  return impl_->Allocate(size, alignment, out);
 }
 
-Status ProxyMemoryPool::Reallocate(int64_t old_size, int64_t new_size, uint8_t** ptr) {
-  return impl_->Reallocate(old_size, new_size, ptr);
+Status ProxyMemoryPool::Reallocate(int64_t old_size, int64_t new_size, int64_t alignment,
+                                   uint8_t** ptr) {
+  return impl_->Reallocate(old_size, new_size, alignment, ptr);
 }
 
-void ProxyMemoryPool::Free(uint8_t* buffer, int64_t size) {
-  return impl_->Free(buffer, size);
+void ProxyMemoryPool::Free(uint8_t* buffer, int64_t size, int64_t alignment) {
+  return impl_->Free(buffer, size, alignment);
 }
 
 int64_t ProxyMemoryPool::bytes_allocated() const { return impl_->bytes_allocated(); }
@@ -802,8 +809,9 @@ std::vector<std::string> SupportedMemoryBackendNames() {
 /// A Buffer whose lifetime is tied to a particular MemoryPool
 class PoolBuffer final : public ResizableBuffer {
  public:
-  explicit PoolBuffer(std::shared_ptr<MemoryManager> mm, MemoryPool* pool)
-      : ResizableBuffer(nullptr, 0, std::move(mm)), pool_(pool) {}
+  explicit PoolBuffer(std::shared_ptr<MemoryManager> mm, MemoryPool* pool,
+                      int64_t alignment)
+      : ResizableBuffer(nullptr, 0, std::move(mm)), pool_(pool), alignment_(alignment) {}
 
   ~PoolBuffer() override {
     // Avoid calling pool_->Free if the global pools are destroyed
@@ -814,7 +822,7 @@ class PoolBuffer final : public ResizableBuffer {
     // no guarantee of destructor order between thread/memory pools)
     uint8_t* ptr = mutable_data();
     if (ptr && !global_state.is_finalizing()) {
-      pool_->Free(ptr, capacity_);
+      pool_->Free(ptr, capacity_, alignment_);
     }
   }
 
@@ -826,9 +834,9 @@ class PoolBuffer final : public ResizableBuffer {
     if (!ptr || capacity > capacity_) {
       int64_t new_capacity = bit_util::RoundUpToMultipleOf64(capacity);
       if (ptr) {
-        RETURN_NOT_OK(pool_->Reallocate(capacity_, new_capacity, &ptr));
+        RETURN_NOT_OK(pool_->Reallocate(capacity_, new_capacity, alignment_, &ptr));
       } else {
-        RETURN_NOT_OK(pool_->Allocate(new_capacity, &ptr));
+        RETURN_NOT_OK(pool_->Allocate(new_capacity, alignment_, &ptr));
       }
       data_ = ptr;
       capacity_ = new_capacity;
@@ -847,7 +855,7 @@ class PoolBuffer final : public ResizableBuffer {
       int64_t new_capacity = bit_util::RoundUpToMultipleOf64(new_size);
       if (capacity_ != new_capacity) {
         // Buffer hasn't got yet the requested size.
-        RETURN_NOT_OK(pool_->Reallocate(capacity_, new_capacity, &ptr));
+        RETURN_NOT_OK(pool_->Reallocate(capacity_, new_capacity, alignment_, &ptr));
         data_ = ptr;
         capacity_ = new_capacity;
       }
@@ -859,7 +867,7 @@ class PoolBuffer final : public ResizableBuffer {
     return Status::OK();
   }
 
-  static std::shared_ptr<PoolBuffer> MakeShared(MemoryPool* pool) {
+  static std::shared_ptr<PoolBuffer> MakeShared(MemoryPool* pool, int64_t alignment) {
     std::shared_ptr<MemoryManager> mm;
     if (pool == nullptr) {
       pool = default_memory_pool();
@@ -867,10 +875,10 @@ class PoolBuffer final : public ResizableBuffer {
     } else {
       mm = CPUDevice::memory_manager(pool);
     }
-    return std::make_shared<PoolBuffer>(std::move(mm), pool);
+    return std::make_shared<PoolBuffer>(std::move(mm), pool, alignment);
   }
 
-  static std::unique_ptr<PoolBuffer> MakeUnique(MemoryPool* pool) {
+  static std::unique_ptr<PoolBuffer> MakeUnique(MemoryPool* pool, int64_t alignment) {
     std::shared_ptr<MemoryManager> mm;
     if (pool == nullptr) {
       pool = default_memory_pool();
@@ -878,11 +886,12 @@ class PoolBuffer final : public ResizableBuffer {
     } else {
       mm = CPUDevice::memory_manager(pool);
     }
-    return std::make_unique<PoolBuffer>(std::move(mm), pool);
+    return std::make_unique<PoolBuffer>(std::move(mm), pool, alignment);
   }
 
  private:
   MemoryPool* pool_;
+  int64_t alignment_;
 };
 
 namespace {
@@ -899,13 +908,26 @@ inline Result<BufferPtr> ResizePoolBuffer(PoolBufferPtr&& buffer, const int64_t
 }  // namespace
 
 Result<std::unique_ptr<Buffer>> AllocateBuffer(const int64_t size, MemoryPool* pool) {
-  return ResizePoolBuffer<std::unique_ptr<Buffer>>(PoolBuffer::MakeUnique(pool), size);
+  return AllocateBuffer(size, kDefaultBufferAlignment, pool);
+}
+
+Result<std::unique_ptr<Buffer>> AllocateBuffer(const int64_t size,
+                                               const int64_t alignment,
+                                               MemoryPool* pool) {
+  return ResizePoolBuffer<std::unique_ptr<Buffer>>(
+      PoolBuffer::MakeUnique(pool, alignment), size);
 }
 
 Result<std::unique_ptr<ResizableBuffer>> AllocateResizableBuffer(const int64_t size,
                                                                  MemoryPool* pool) {
-  return ResizePoolBuffer<std::unique_ptr<ResizableBuffer>>(PoolBuffer::MakeUnique(pool),
-                                                            size);
+  return AllocateResizableBuffer(size, kDefaultBufferAlignment, pool);
+}
+
+Result<std::unique_ptr<ResizableBuffer>> AllocateResizableBuffer(const int64_t size,
+                                                                 const int64_t alignment,
+                                                                 MemoryPool* pool) {
+  return ResizePoolBuffer<std::unique_ptr<ResizableBuffer>>(
+      PoolBuffer::MakeUnique(pool, alignment), size);
 }
 
 }  // namespace arrow
