@@ -61,10 +61,10 @@ quietly <- !env_is("ARROW_R_DEV", "true")
 # and don't fall back to a full source build
 build_ok <- !env_is("LIBARROW_BUILD", "false")
 
-# Check if we're doing an offline build.
+# Check if we're authorized to download (not asked an offline build).
 # (Note that cmake will still be downloaded if necessary
 #  https://arrow.apache.org/docs/developers/cpp/building.html#offline-builds)
-download_ok <- !test_mode && !env_is("TEST_OFFLINE_BUILD", "true") && try_download("https://raw.githubusercontent.com/apache/arrow/master/r/DESCRIPTION", tempfile(), hush = TRUE)
+download_ok <- !test_mode && !env_is("TEST_OFFLINE_BUILD", "true")
 
 # This "tools/thirdparty_dependencies" path, within the tar file, might exist if
 # create_package_with_all_dependencies() was run, or if someone has created it
@@ -83,7 +83,8 @@ download_binary <- function(lib) {
     }
   } else {
     if (!quietly) {
-      cat(sprintf("*** No libarrow binary found for version %s (%s)\n", VERSION, lib))
+      cat(sprintf("*** Downloading libarrow binary failed for version %s (%s)\n    at %s\n",
+                  VERSION, lib, binary_url))
     }
     libfile <- NULL
   }
@@ -438,8 +439,8 @@ build_libarrow <- function(src_dir, dst_dir) {
     cat(paste0(
       "*** Building C++ library from source, but downloading thirdparty dependencies\n",
       "    is not possible, so this build will turn off all thirdparty features.\n",
-      "    See install vignette for details:\n",
-      "    https://cran.r-project.org/web/packages/arrow/vignettes/install.html\n"
+      "    See installation guide for details:\n",
+      "    https://arrow.apache.org/docs/r/articles/install.html\n"
     ))
     env_var_list <- turn_off_all_optional_features(env_var_list)
   } else if (dir.exists(thirdparty_dependency_dir)) {
@@ -706,13 +707,25 @@ if (!test_mode && !file.exists(paste0(dst_dir, "/include/arrow/api.h"))) {
   # don't need to do anything. Otherwise,
   # (1) Look for a prebuilt binary for this version
   bin_file <- src_dir <- NULL
-  if (download_ok) {
+
+  if (!identical(Sys.getenv("ARROW_DOWNLOADED_BINARIES"), "")) {
+    bin_zip <- Sys.getenv("ARROW_DOWNLOADED_BINARIES")
+    cat(sprintf("*** Using pre-downloaded zip for libarrow binaries: %s\n", bin_zip))
+    if (file.exists(bin_zip)) {
+      bin_file <- tempfile()
+      file.copy(bin_zip, bin_file)
+    } else {
+      cat(sprintf("*** File not found: %s ($ARROW_DOWNLOADED_BINARIES)\n", bin_zip))
+      bin_file <- NULL
+    }
+  } else if (download_ok) {
     binary_flavor <- identify_binary()
     if (!is.null(binary_flavor)) {
       # The env vars say we can, and we've determined a lib that should work
       bin_file <- download_binary(binary_flavor)
     }
   }
+
   if (!is.null(bin_file)) {
     # Extract them
     dir.create(dst_dir, showWarnings = !quietly, recursive = TRUE)
@@ -724,14 +737,14 @@ if (!test_mode && !file.exists(paste0(dst_dir, "/include/arrow/api.h"))) {
     if (!is.null(src_dir)) {
       cat(paste0(
         "*** Building libarrow from source\n",
-        "    For build options and troubleshooting, see the install vignette:\n",
-        "    https://cran.r-project.org/web/packages/arrow/vignettes/install.html\n"
+        "    For build options and troubleshooting, see the install guide:\n",
+        "    https://arrow.apache.org/docs/r/articles/install.html\n"
       ))
       build_libarrow(src_dir, dst_dir)
     } else {
-      cat("*** Proceeding without libarrow\n")
+      cat("*** Proceeding without libarrow (no local source)\n")
     }
   } else {
-    cat("*** Proceeding without libarrow\n")
+    cat("*** Proceeding without libarrow (build not authorized)\n")
   }
 }
